
import datetime as dt
import glob
import math
import numpy as np
from osgeo import gdal
import scipy.interpolate

import biota

import pdb
    


def _readTimeStack(tile, date, search_days = 7):
    '''
    Load soil moisture time stack
    '''
    
    # Load list of netCDF files
    nc_files = np.array(sorted(glob.glob(tile.SM_dir + '/*.nc')))
    
    assert len(nc_files) > 0, "No data found in soil moisture data path (%s)."%nc_files
    
    # Grab their dates
    nc_dates = np.array([dt.datetime.strptime(nc_file.split('/')[-1].split('-')[-2][:8], '%Y%m%d').date() for nc_file in nc_files])

    sel = np.logical_and(nc_dates >= date - dt.timedelta(search_days), nc_dates <= date + dt.timedelta(search_days))

    assert len(nc_files[sel]) > 0, "No data for tile date found in soil moisture data path (%s)."%nc_files

    for n, nc_file in enumerate(nc_files[sel]):
            
        # Load dataset
        ds = gdal.Open('NETCDF:%s:sm'%nc_file)
        
        # Get GeoTransform()
        geo_t = ds.GetGeoTransform()
        
        # Get netCDF extent
        ulLon, ulLat = geo_t[0], geo_t[3]
        lonDist, latDist = geo_t[1], geo_t[5]
        
        # Get UL coord for tile
        xOffset = int(math.floor((tile.lon - ulLon) / lonDist))
        yOffset = int(math.ceil((ulLat - tile.lat) / latDist)) * -1
        
        # And the extent of pixels to load
        yCount = int(abs(math.floor(1. / latDist)))
        xCount = int(math.ceil(1. / lonDist))
        
        this_sm = ds.ReadAsArray(xOffset, yOffset, xCount, yCount)
        
        if 'sm_out' not in locals():
            sm_out = np.ma.zeros((this_sm.shape[0], this_sm.shape[1], (search_days * 2) + 1), dtype = np.float32)
        
        sm_out[:,:,n] = np.ma.array(this_sm, mask = this_sm == -9999.)
            
    geo_t_out = (ulLon + (xOffset * lonDist), geo_t[1], geo_t[2], ulLat + (yOffset * latDist), geo_t[4], geo_t[5])
    
    return sm_out, geo_t_out


def _interpolateTime(sm):
    '''
    '''

    # Interpolate through time
    sm_interp = np.zeros((sm.shape[0], sm.shape[1]), dtype = np.float32)
    
    for i in range(sm.shape[0]):
        for j in range(sm.shape[1]):
            
            this_sm = sm[i,j,:]
            x = np.arange(sm.shape[-1])
            y = this_sm.data
            
            if np.sum(this_sm.mask == False) > 1:
                f = scipy.interpolate.interp1d(x[this_sm.mask == False], y[this_sm.mask == False], fill_value = 'extrapolate')
                interp = f(x)[(x[this_sm.mask == False]).shape[0]/2]
            else:
                interp = -9999.
            
            sm_interp[i, j] = interp
            
    return sm_interp

def _resampleSM(sm_interp, tile, geo_t):
    '''
    '''
        
    # Create output file matching ALOS tile
    gdal_driver = gdal.GetDriverByName('MEM')
    
    ds_source = gdal_driver.Create('', sm_interp.shape[0], sm_interp.shape[1], 1, gdal.GDT_Float32)
    ds_source.SetGeoTransform(geo_t)
    ds_source.SetProjection(tile.proj)
    ds_source.GetRasterBand(1).WriteArray(sm_interp)
    
    ds_dest = gdal_driver.Create('', tile.ySize, tile.xSize, 1, gdal.GDT_Float32)
    ds_dest.SetGeoTransform(tile.geo_t)
    ds_dest.SetProjection(tile.proj)
    
    # Reproject input GeoTiff to match the ALOS tile
    gdal.ReprojectImage(ds_source, ds_dest, tile.proj, tile.proj, gdal.GRA_NearestNeighbour)
    
    # Load resampled image into memory
    sm_resampled = ds_dest.GetRasterBand(1).ReadAsArray()
    
    return np.ma.array(sm_resampled, mask = sm_resampled == -9999.)


def _loadSM(tile, date, search_days = 7):
    '''
    Load and reproject soil moisture data
    '''
    
    # Load SM time stack (within 1 week of measurement)
    sm, geo_t = _readTimeStack(tile, date, search_days = search_days)
    
    # Interpolate over time gaps
    sm_interp = _interpolateTime(sm)
    
    # Reproject to match ALOS tile (nearest neighbor)
    sm_resampled = _resampleSM(sm_interp, tile, geo_t)
    
    return sm_resampled

            
def getSM(tile, search_days = 7):
    """
    Function to load a resampled soil moisture image from the ESA CCI soil moisture product. The function returns a masked array with estimated volumetric soil moisture content (m^2/m^2), averaged for each satellite overpass in a tile.
    
    Args:
        tile: A biota.LoadTile() object.
        
    Returns:
        A masked array of estimated soil moisture (m^2/m^2)
    """
        
    # Build an array of dates in string format, which is quicker to search
    dates = tile.getDate().astype(np.int)
    
    # Generate output array
    out = np.zeros_like(tile.mask, dtype = np.float32) + 999999.
    
    for date in np.unique(dates): #np.unique(tile.getDate()).astype(np.int):        
        
        # Nodata (0 = 1st Jan 1970)
        if date == 0:
            continue
        
        # Interpolate through time and draw out the central value, and zoom to scale of tile
        sm_resampled = _loadSM(tile, dt.datetime(1970,1,1).date() + dt.timedelta(date), search_days = search_days)
        
        # Only output soil moisture where > 25 % data exists        
        if float((sm_resampled.mask[dates == date]).sum()) / (sm_resampled.mask[dates == date]).shape[0] <= 0.25:
            out[dates == date] = np.ma.mean(sm_resampled[dates == date])        
    
    # Add the mask
    out = np.ma.array(out, mask = np.logical_or(tile.mask, out == 999999.))
        
    return out
    
            


#!/usr/bin/env python

import argparse
import datetime
import ftplib
import math
import os
import subprocess
import tarfile

import pdb

"""
This is a simple script to assist in the downloading of data from the ALOS mosaic product.
"""


def generateURL(lat, lon, year, large_tile = False):
    """
    Generates a URL to fetch ALOS mosaic data from JAXA FTP server.
    """
    
    # Test that inputs are reasonable lats/lons
    if large_tile:
        assert lat % 5 == 0, "Latitudes must be a multiple of 5 degrees if downloading a 5x5 degree tile."
        assert lon % 5 == 0, "Longitudes must be a multiple of 5 degrees if downloading a 5x5 degree tile."
    
    # Test that years are reasonable
    assert (year >= 2007 and year <= 2010) or (year >= 2015 and year <= datetime.datetime.now().year), "Years must be in the range 2007 - 2010 and 2015 - present. Your year was %s."%str(year)
    
    # Test that lat/lon is reasonable
    assert (lat >= -90 and lat <= 90) and (lon >= -180 and lon <= 180), "Latitudes must be between -90 and 90 degrees, and lontitudes between -180 and 180 degrees."
    
    # Get hemisphere
    hem_NS = 'S' if lat < 0 else 'N'
    hem_EW = 'W' if lon < 0 else 'E'
    
    tile_name = '%s%s%s%s'%(hem_NS, str(abs(lat)).zfill(2), hem_EW, str(abs(lon)).zfill(3))

    # Get URL for a 5x5 tile
    if large_tile:
        
        # Filename patterns are different for ALOS-1/ALOS-2
        if year <= 2010:
            url = 'ftp://ftp.eorc.jaxa.jp/pub/ALOS/ext1/PALSAR_MSC/25m_MSC/%s/%s_%s_MOS.tar.gz'
        elif year > 2010 and year < 2017:
            url = 'ftp://ftp.eorc.jaxa.jp/pub/ALOS-2/ext1/PALSAR-2_MSC/25m_MSC/%s/%s_%s_MOS_F02DAR.tar.gz'
        else:
            url = 'ftp://ftp.eorc.jaxa.jp/pub/ALOS-2/ext2/PALSAR-2_MSC/25m_MSC/%s/%s_%s_MOS_F02DAR.tar.gz'	  
                
        url = url%(str(year), tile_name, str(year)[-2:])

    # Get URL for a 1x1 tile
    else:
        
        # Filename patterns are different for ALOS-1/ALOS-2
        if year <= 2010:
            url = 'ftp://ftp.eorc.jaxa.jp/pub/ALOS/ext1/PALSAR_MSC/25m_MSC/%s/%s/%s_%s_MOS.tar.gz'
        else:
            url = 'ftp://ftp.eorc.jaxa.jp/pub/ALOS-2/ext1/PALSAR-2_MSC/25m_MSC/%s/%s/%s_%s_MOS_F02DAR.tar.gz'
        
        # Special consideration for directory name of 1x1 tile       
        lat_dir = int(5 * math.ceil(float(lat)/5))
        lon_dir = int(5 * math.floor(float(lon)/5))
        hem_NS_dir = 'S' if lat_dir < 0 else 'N'
        hem_EW_dir = 'W' if lon_dir < 0 else 'E'
        
        directory_name = '%s%s%s%s'%(hem_NS_dir, str(abs(lat_dir)).zfill(2), hem_EW_dir, str(abs(lon_dir)).zfill(3))
        
        url = url%(str(year), directory_name, tile_name, str(year)[-2:])
        
    return url


def download(url, output_dir = os.getcwd()):
    """
    Download data from JAXA FTP server
    """
    
    # Check that output directory exists
    assert os.path.isdir(output_dir), "The output directory (%s) does not exist. Create it, then try again."%output_dir
    
    # Check that output file doesn't already exist
    this_file = '%s/%s'%(output_dir,url.split('/')[-1])
    
    if os.path.exists(this_file):
        print "WARNING: File %s already exists. Skipping."%this_file
    elif os.path.exists(this_file[:-7]):
        print "WARNING: File %s already exists. Skipping."%this_file[:-7]

    else:
        # Download
        exit_status = subprocess.call(['wget', '-nc', url, '-P', output_dir])
        
        if exit_status == 8:
            raise ValueError('Download failed: a tile for that location was not found on the FTP server.')
        if exit_status != 0:
            raise ValueError('Download failed with wget error code %s.'%str(exit_status))
    
    # Determine absolute path of downloaded file
    filepath = '%s/%s'%(output_dir.rstrip('/'), url.split('/')[-1])
    
    return filepath


def decompress(targz_file, remove = False):
    '''
    Unzips .tar.gz ALOS mosaic files downloaded from JAXA, and removes original where requested.
    '''
    
    assert targz_file.endswith("tar.gz"), "File name must end with .tar.gz to be decompressed."
    
    # Check that output file doesn't already exist
    if os.path.exists(targz_file.split('/')[-1].split('.')[0]):
        print 'WARNING: File %s already exists at output location. Not extracting.'%targz_file
    
    else:
        print 'Extracting %s'%targz_file
            
        tar = tarfile.open(targz_file, "r:gz")
        tar.extractall(path = targz_file[:-7])
        tar.close()
        
        if remove: removeTarGz(targz_file)


def removeTarGz(targz_file):
    """
    Deletes ALOS-1/ALOS-2 .tar.gx files from disk.
    Input is a compress ALOS-1/ALOS-2 file from JAXA.
    """
    
    assert targz_file.endswith('_MOS.tar.gz') or targz_file.endswith('_MOS_F02DAR.tar.gz'), "removeTarGz function should only be used to delete ALOS-1/ALOS2 .tar.gz files"
    
    os.remove(targz_file)


def checkYears(years):
    """
    Reduces input years to those available for the ALOS mosaic (or may be available in future).
    """
    
    years_cleansed = []
    
    for y in years:
        
        # Remove years before ALOS-1
        if y < 2007:   
            raise ValueError("Can't download data for year %s; no data from ALOS are available before 2007."%str(y))
            
        # Remove years between ALOS-1 and ALOS-2
        elif y > 2010 and y < 2015:
            raise ValueError("Can't download data for year %s; no data from ALOS are available 2011 to 2014 (inclusive)."%str(y))
        
        # Remove years from the future, which can't possibly exist yet.
        elif y > datetime.datetime.now().year:
            raise ValueError("Can't download data for %s; this year is in the future."%str(y))
        


def main(lat, lon, year, large_tile = False, output_dir = os.getcwd(), remove = False):
    '''
    Run through data download and preparation chain
    '''
    
    # Generate download URL
    url = generateURL(lat, lon, year, large_tile = large_tile)
    
    # Get file, sending it to output_dir
    filepath = download(url, output_dir = output_dir)
    
    # Decompress downloaded file, and delete original if remove == True
    decompress(filepath, remove = remove)

    

if __name__ == '__main__':

    # Set up command line parser
    parser = argparse.ArgumentParser(description = 'Download ALOS-1/2 data from JAXA, specifying a particular year and latitude/longitude.')
    
    parser._action_groups.pop()
    required = parser.add_argument_group('Required arguments')
    optional = parser.add_argument_group('Optional arguments')
    
    # Required arguments
    required.add_argument('-lat', '--latitude', metavar = 'DEG', type = int, help = "Latitude of tile upper-left corner.")
    required.add_argument('-lon', '--longitude', metavar = 'DEG', type = int, help = "Longitude of tile upper-left corner.")
    
    # Optional arguments
    optional.add_argument('-l', '--large', action = 'store_true', default = False, help = "Download large tiles. ALOS mosaic tiles are available in 1x1 or 5x5 degree tiles. If downloading large volumes of data, it's usually better to use the latter. If this option is chosen, you must select a lat and lon that's a multiple of 5 degrees.")
    optional.add_argument('-y', '--years', metavar = 'Y', type = int, nargs = '+', default = [2007, 2008, 2009, 2010, 2015, 2016], help = "Year of data to download. Defaults to downloading all data.")
    optional.add_argument('-o', '--output_dir',metavar = 'DIR',  type = str, default = os.getcwd(), help = "Optionally specify an output directory. Defaults to the present working directory.")
    optional.add_argument('-r', '--remove', action='store_true', default = False, help = "Optionally remove downloaded .tar.gz files after decompression.")

    # Get arguments from command line
    args = parser.parse_args()
    
    # Cleanse input years
    checkYears(args.years)
    
    # Run through entire processing sequence
    for year in args.years:
        
        main(args.latitude, args.longitude, year, large_tile = args.large, output_dir = args.output_dir, remove = args.remove)




import matplotlib.pyplot as plt
import numpy as np
import os


import pdb

def loadArray(filepath):
    """
    Use gdal to load a geospatial image into a numpy array
    """
    
    from osgeo import gdal
    
    ds = gdal.Open(filepath, 0)
    
    return ds.ReadAsArray()
    
    
def loadGeoTransform(filepath):
    """
    Use gdal to load a gdal geotransform (affine transform) tuple
    """
    
    from osgeo import gdal
        
    ds = gdal.Open(filepath, 0)
    
    return ds.GetGeoTransform()


def loadProjection(filepath):
    """
    Use gdal to load projection info
    """
    
    from osgeo import gdal
    
    ds = gdal.Open(filepath, 0)
    
    return ds.GetProjection()


def loadSize(filepath):
    """
    Use gdal to load raster
    """
    
    from osgeo import gdal
    
    ds = gdal.Open(filepath, 0)
    
    return ds.RasterYSize, ds.RasterXSize


def loadRaster(raster, tile, resampling = 0):
    """
    Loads a raster image (e.g. GeoTiff), reprojecting to match a tile.
    
    Args:
        raster: path to raster file (string)
        tile: A tile from biota.LoadTile()
        resampling: Optionally specify a gdal resampling type. Defaults to gdal.GRA_NearestNeighbour.
        
    Returns:
        A numpy array matching tile
    """
    
    from osgeo import gdal
    
    # Open GeoTiff and get metadata
    ds_source = gdal.Open(raster)
    proj_source = loadProjection(raster)
    
    # Create output file matching ALOS tile
    gdal_driver = gdal.GetDriverByName('MEM')
    ds_dest = gdal_driver.Create('', tile.xSize, tile.ySize, 1, 3)
    ds_dest.SetGeoTransform(tile.geo_t)
    ds_dest.SetProjection(tile.proj)
       
    # Reproject input GeoTiff to match the ALOS tile
    gdal.ReprojectImage(ds_source, ds_dest, proj_source, tile.proj, resampling)
    
    # Load resampled image into memory
    resampled = ds_dest.GetRasterBand(1).ReadAsArray()
    
    return resampled

def outputGeoTiff(data, filename, geo_t, proj, output_dir = os.getcwd(), dtype = 6, nodata = None):
    """
    Writes a GeoTiff file to disk.
    
    Args:
        data: A numpy array.
        geo_t: A GDAL geoMatrix (ds.GetGeoTransform()).
        proj: A GDAL projection (ds.GetProjection()).
        filename: Specify an output file name.
        output_dir: Optioanlly specify an output directory. Defaults to working directory.
        dtype: gdal data type (gdal.GDT_*). Defaults to gdal.GDT_Float32.
        nodata: The nodata value for the array
    """
    
    from osgeo import osr, gdal
    
    # Get full output path
    output_path = '%s/%s.tif'%(os.path.abspath(os.path.expanduser(output_dir)), filename.rstrip('.tif'))
    
    # Save image with georeference info
    driver = gdal.GetDriverByName('GTiff')
    ds = driver.Create(output_path, data.shape[1], data.shape[0], 1, dtype, options = ['COMPRESS=LZW'])
    ds.SetGeoTransform(geo_t)
    ds.SetProjection(proj)
        
    # Set nodata
    if nodata != None:
        ds.GetRasterBand(1).SetNoDataValue(nodata)
    
    # Write data for masked and unmasked arrays
    if np.ma.isMaskedArray(data):
        ds.GetRasterBand(1).WriteArray(data.filled(nodata))
    else:
        ds.GetRasterBand(1).WriteArray(data)
    ds = None
    


def buildMap(fig, ax, data, lat, lon, title ='', cbartitle = '', vmin = None, vmax = None, cmap = None, big_labels = False):
    """
    Builds a standardised map for overviewFigure() and showFigure().
    """
    
    labelsize = 5
    titlesize = 8
    cbarlabelsize = 6
    cbartitlesize = 7
    
    if big_labels:
        labelsize *= 2
        titlesize *= 2
        cbarlabelsize *= 2
        cbartitlesize *= 2
    
        
    im = ax.imshow(data, vmin = vmin, vmax = vmax, cmap = cmap, interpolation = 'nearest', aspect = 'auto')
    
    ax.set_yticks(np.arange(0,data.shape[0],data.shape[0]/10))
    ax.set_xticks(np.arange(0,data.shape[1],data.shape[1]/10))
    ax.set_yticklabels(["%.1f" %i for i in np.arange(lat, lat - 1.01, - 0.1)])
    ax.set_xticklabels(["%.1f" %i for i in np.arange(lon, lon + 1.01, 0.1)])
    ax.tick_params(labelsize = labelsize)
    ax.set_ylabel('Latitude', fontsize = labelsize)
    ax.set_xlabel('Longitude', fontsize = labelsize)
    ax.set_title(title, fontsize = titlesize)
    
    cbar = fig.colorbar(im, ax = ax, fraction = 0.046, pad = 0.04)
    cbar.ax.tick_params(labelsize = cbarlabelsize)
    cbar.set_label(cbartitle, fontsize = cbartitlesize)


def showFigure(data, lat, lon, title = None, cbartitle = None, vmin = None, vmax = None, cmap = None, big_labels = True):
    '''
    Show an overview image.
    
    Args:
        data: A numpy array containing the data to display
        lat: Latitude, an integer
        lon: Longitude, an integer
        title: A string to show as the figure title
        cbartitle: A string to show as the colorbar title
        vmin: Min colorbar range
        vmax: Max colorbar range
        cmap: String of a matplotlib colorbar
    '''
    
    # Set up new figure
    fig = plt.figure(figsize = (7, 6))
    ax = fig.add_subplot(1, 1, 1)
    
    # Plot map
    buildMap(fig, ax, data, lat, lon, title = title, cbartitle = cbartitle, vmin = vmin, vmax = vmax, cmap = cmap, big_labels = big_labels)
    
    # Show in window
    plt.show()
        
    # Tidy up
    plt.close()
        

def overviewFigure(tile_change, output = False, show = True):
    """
    Generate an overview image showing biomass and proportional biomass change for an ALOS tile.
    
    Args:
        tile_change: An ALOS change object from biota.LoadChange()
        output: Set to True to output a summary png image
        show: Set to True to display image on screen
    """
    
    import matplotlib.pyplot as plt
    
    # Load AGB, and update masks to exclude areas outisde forest definition. Good for visualisation
    AGB_t1 = tile_change.tile_t1.getAGB()
    AGB_t2 = tile_change.tile_t2.getAGB()
    
    # Mask out areas < 10 tC/ha
    AGB_t1 = np.ma.array(AGB_t1, mask = np.logical_or(AGB_t1.mask, AGB_t1 < 10.))
    AGB_t2 = np.ma.array(AGB_t2, mask = np.logical_or(AGB_t2.mask, AGB_t1 < 10.))
        
    AGB_change = AGB_t2 - AGB_t1
    #AGB_pcChange = 100 * (AGB_change / AGB_t1) # %
    
    # Calculate change type
    change_type = tile_change.getChangeType()
    change_code = tile_change.ChangeCode
    
    # Set minor loss and minor gain to nodata
    change_code[np.logical_or(change_code == 3, change_code == 4)] = 0
    change_code.mask[change_type.data == 0] = True
    
    fig = plt.figure(figsize = (7, 6))
    
    # Plot a map of AGB at t1
    ax1 = fig.add_subplot(2, 2, 1)
    buildMap(fig, ax1, AGB_t1, tile_change.lat, tile_change.lon, title = 'AGB %s'%str(tile_change.year_t1), cbartitle = 'tC/ha', vmin = 10., vmax = 40., cmap = 'YlGn')
    
    # Plot a map of AGB at t2
    ax2 = fig.add_subplot(2, 2, 2, sharex = ax1, sharey = ax1)
    buildMap(fig, ax2, AGB_t2, tile_change.lat, tile_change.lon, title = 'AGB %s'%str(tile_change.year_t2), cbartitle = 'tC/ha', vmin = 10., vmax = 40., cmap = 'YlGn')    
    
    # Plot a map of absolute AGB change   
    ax3 = fig.add_subplot(2, 2, 3, sharex = ax1, sharey = ax1)
    buildMap(fig, ax3, AGB_change, tile_change.lat, tile_change.lon, title = 'AGB change (%s-%s)'%(str(tile_change.tile_t1.year),str(tile_change.year_t2)),
              cbartitle = 'tC/ha', vmin = -10., vmax = 10., cmap = 'RdBu')    
    
    # Plot a map of % AGB change
    ax4 = fig.add_subplot(2, 2, 4, sharex = ax1, sharey = ax1)
    buildMap(fig, ax4, change_code, tile_change.lat, tile_change.lon, title = 'Change type (%s-%s)'%(str(tile_change.tile_t1.year),str(tile_change.year_t2)),
              vmin = 1., vmax = 6., cmap = 'Spectral')
    
    plt.tight_layout()
    
    # Output image to png
    if output:
        output_pattern = tile_change.output_pattern.replace('.tif','.png')
    
        output_path = os.path.abspath(os.path.expanduser('%s/%s'%(tile_change.output_dir, output_pattern%('OverviewFigure'))))
    
        plt.savefig(output_path, dpi = 150)
    
    # Display image on screen
    if show:
        plt.show()
        
    plt.close()





#!/usr/bin/env python

import argparse
import datetime as dt
import itertools
import math
import numpy as np
import os
from scipy import ndimage
import scipy.stats as stats
import skimage.measure

import matplotlib.pyplot as plt
import pdb

import biota.filter
import biota.indices
import biota.IO
import biota.mask
import biota.SM

"""
These are two classes for loading individual tiles and compliling them into a change object.
"""



class LoadTile(object):
    """
    An ALOS mosaic tile.
    Mosaic tiles have the following properties:

    Attributes:
        lat:
        lon:
        DN: An array of uncalibrated digital numbers from the ALOS tile.
        mask:
    """
        
    def __init__(self, data_dir, lat, lon, year, forest_threshold = 10., area_threshold = 0., downsample_factor = 1, lee_filter = False, sm_dir = os.getcwd(), output_dir = os.getcwd()):
        """
        Loads data and metadata for an ALOS mosaic tile.
        """
        
        from osgeo import gdal
        
        # Test that inputs are of reasonable lats/lons/years
        assert type(lat) == int, "Latitude must be an integer."
        assert lat < 90. or lat > -90., "Latitude must be between -90 and 90 degrees."
        assert type(lon) == int, "Longitude must be an integer."
        assert lon < 180. or lon > -180., "Longitude must be between -180 and 180 degrees."
        assert type(year) == int, "Year must be an integer."
        assert (year >= 2007 and year <= 2010) or (year >= 2015 and year <= dt.datetime.now().year), "Years must be in the range 2007 - 2010 and 2015 - present. Your input year was %s."%str(year)
        assert downsample_factor >= 1 and type(downsample_factor) == int, "Downsampling factor must be an integer greater than 1."
        assert type(lee_filter) == bool, "Option lee_filter must be set to 'True' or 'False'."
        assert os.path.isdir(os.path.expanduser(data_dir)), "Specified data directory (%s) does not exist"%str(data_dir)
        assert os.path.isdir(os.path.expanduser(sm_dir)), "Specified soil moisture directory (%s) does not exist"%str(sm_dir)
        assert os.path.isdir(os.path.expanduser(output_dir)), "Specified output directory (%s) does not exist"%str(output_dir)
        assert type(forest_threshold) == float or type(forest_threshold) == int, "Forest threshold must be numeric."
        assert type(area_threshold) == float or type(area_threshold) == int, "Area threshold must be numeric."
        
        self.lat = lat
        self.lon = lon
        self.year = year
        self.downsample_factor = downsample_factor
        self.lee_filter = lee_filter
        self.forest_threshold = forest_threshold
        self.area_threshold = area_threshold
        
        # Deterine hemispheres
        self.hem_NS = 'S' if lat < 0 else 'N'
        self.hem_EW = 'W' if lon < 0 else 'E'
        
        # Determine whether ALOS-1 or ALOS-2
        self.satellite = self.__getSatellite()
        
        # Determine filenames
        self.data_dir = os.path.expanduser(data_dir.rstrip('/'))
        self.directory = self.__getDirectory()
        self.HH_path = self.__getHHPath()
        self.HV_path = self.__getHVPath()
        self.mask_path = self.__getMaskPath()
        self.date_path = self.__getDatePath()
        
        # Set up soil moisture data location
        self.SM_dir = os.path.expanduser(sm_dir.rstrip('/'))
        
        # Set up locations for file output
        self.output_pattern = self.__getOutputPattern()
        self.output_dir = os.path.expanduser(output_dir.rstrip('/'))

        # Get Raster size
        self.ySize, self.xSize = self.__getSize(self.HH_path)
                
        # Get GDAL geotransform and projection
        self.geo_t = self.__getGeoT()
        self.proj = self.__getProj()
        
        # Get the raster extent
        self.extent = self.__getExtent()
        
        # Determine x and y resolution in meters
        self.xRes = self.__getXRes()
        self.yRes = self.__getYRes()
        
        # Record the nodata value
        self.nodata = self.__getNodata()        
        self.nodata_byte = self.__getNodata(dtype = gdal.GDT_Byte)
        
        # Get Equivalent Number of Looks
        self.nLooks = self.__getnLooks()
        
        # Update image metadata where a degree of resampling is included
        if self.downsample_factor != 1:
            self.__rebin()
        
        # Load mask
        self.mask = self.__getMask()
                
        
    def __getSatellite(self):
        """
        Return the sensor for the ALOS mosaic tile
        """
        
        if self.year >= 2015:
            satellite = 'ALOS-2'
        else:
            satellite = 'ALOS-1'
        
        return satellite
    
    def __getDirectory(self):
        """
        Return the directory containing ALOS data for a given lat/lon. Assumes its distributed as a 5x5 tile at present.
        """
        
        assert os.path.isdir(self.data_dir), "Data location must be a directory."
        
        # Directories and files have standardised pattern
        name_pattern = '%s%s_%s_%s'
        
        # Directory name patterns name patterns are different for ALOS-1/ALOS-2
        if self.satellite == 'ALOS-2':
            name_pattern += '_F02DAR'
        
        # First, check if the 1x1 tile is present at data_dir
        lat_dir = self.hem_NS + str(abs(self.lat)).zfill(2)
        lon_dir = self.hem_EW + str(abs(self.lon)).zfill(3)
        
        # Generate directory name
        directory = self.data_dir + '/' + name_pattern%(lat_dir, lon_dir, str(self.year)[-2:], 'MOS') + '/'
        
        # If no 1x1 tile, check that 5x5 tile exists
        if not os.path.isdir(directory):
                        
            # Calculate the hemisphere and lat/lon of the 5x5 tile
            lat_large = self.lat + (5 - self.lat) % 5
            lon_large = self.lon - (self.lon % 5)
            hem_NS_large = 'S' if lat_large < 0 else 'N'
            hem_EW_large = 'W' if lon_large < 0 else 'E'
            
            # Get lat/lon for directory (lat/lon of upper-left hand corner of 5x5 tiles)
            lat_dir = hem_NS_large + str(abs(lat_large)).zfill(2)
            lon_dir = hem_EW_large + str(abs(lon_large)).zfill(3)
        
            # Generate directory name
            directory = self.data_dir + '/' + name_pattern%(lat_dir, lon_dir, str(self.year)[-2:], 'MOS') + '/'
            
            if not os.path.isdir(directory):
                raise IOError('No tile for lat: %s, lon: %s exists in the specified data directory.'%(str(self.lat), str(self.lon)))
        
        return directory
    
    def __getFilename(self, append_pattern):
        """
        Return the filename for ALOS data for a given lat/lon.
        """
      
        # Get lat/lon for filename
        lat_file = self.hem_NS + str(abs(self.lat)).zfill(2)
        lon_file = self.hem_EW + str(abs(self.lon)).zfill(3)
        
        name_pattern = '%s%s_%s_%s'
        
        # Directory name patterns name patterns are different for ALOS-1/ALOS-2
        if self.satellite == 'ALOS-2':
            name_pattern += '_F02DAR'

        # Generate file name
        return name_pattern%(lat_file, lon_file, str(self.year)[-2:], append_pattern)
    
    def __getHHPath(self):
        """
        Determines the filepath to HV data.
        """
        
        HHpath = self.__getDirectory() + self.__getFilename('sl_HH')
        
        # Stop if the ALOS data don't exist
        if not os.path.isfile(HHpath): 
            raise IOError('No data found for HH polarisation for lat: %s, lon: %s.')
        
        return HHpath
    
    def __getHVPath(self):
        """
        Determines the filepath to HV data.
        """
        
        HVpath = self.__getDirectory() + self.__getFilename('sl_HV')
        
        # Stop if the ALOS data don't exist
        if not os.path.isfile(HVpath): 
            raise IOError('No data found for HV polarisation for lat: %s, lon: %s.')
        
        return HVpath
    
    
    def __getMaskPath(self):
        """
        Determines the filepath to mask data.
        """
                
        mask = self.__getDirectory() + self.__getFilename('mask')
        
        # Stop if the ALOS data don't exist
        if not os.path.isfile(mask): 
            raise IOError('No data found for mask for lat: %s, lon: %s.')
        
        return mask 
    
    def __getDatePath(self):
        """
        Determines the filepath to DOY data.
        """
        
        date = self.__getDirectory() + self.__getFilename('date')
        
        # Stop if the ALOS data don't exist
        if not os.path.isfile(date): 
            raise IOError('No data found for date for lat: %s, lon: %s.')
        
        return date 
    
    def __getOutputPattern(self):
        """
        Generates a filename pattern for data output.
        """
        
        return '%s_%s_%s%s.tif'%('%s', str(self.year), self.hem_NS + str(abs(self.lat)).zfill(2), self.hem_EW + str(abs(self.lon)).zfill(3))
    
    def __getNodata(self, dtype = 6):
        """
        Return nodata values
        """
        
        from osgeo import gdal
        
        # Generate a nodata value
        if dtype == gdal.GDT_Byte:
            nodata = 255
        else:
            nodata = 999999
        
        return nodata
    
    def __getnLooks(self)    :
        """
        Define number of looks. Out of the box, this is 16 for the ALOS mosaic.
        """
        
        return 16

    def __getSize(self, filepath):
        """
        Determines the size of a tile.
        """
                
        y_size, x_size = biota.IO.loadSize(filepath)
        
        return y_size, x_size
    
    def __getGeoT(self):
        """
        Generates a GDAL GeoTransform for tile.
        """
        
        geo_t = (float(self.lon), 1./self.xSize, 0., float(self.lat), 0., -1./self.xSize)
               
        return geo_t
    
    def __getProj(self):
        """
        Fetches projection info for tile.
        """
        
        proj = biota.IO.loadProjection(self.__getHHPath())
        
        return proj
    
    def __getExtent(self):
        """
        Calculates the raster exent in the format [xmin. ymin, xmax, ymax] using the geoTransform and raster size.
        """
        
        xmin = self.geo_t[0]
        ymax = self.geo_t[3]
        xmax = self.geo_t[0] + (self.geo_t[1] * self.xSize)
        ymin = self.geo_t[3] + (self.geo_t[5] * self.ySize)
        
        return (xmin, ymin, xmax, ymax)    
    
    def __getXRes(self):
        """
        Approximates the resolution of a pixel in meters.
        From: https://en.wikipedia.org/wiki/Geographic_coordinate_system
        """
        
        lat = math.radians(self.lat - 0.5)
                
        m_per_degree = (111412.84 * math.cos(lat)) - (93.5 * math.cos(3 * lat)) + (0.118 * math.cos(5 * lat))
        
        return round(m_per_degree / self.xSize, 3)
        
    def __getYRes(self):
        """
        Approximates the resolution of a pixel in meters.
        From: https://en.wikipedia.org/wiki/Geographic_coordinate_system
        """

        lat = math.radians(self.lat - 0.5)
        
        m_per_degree = 111132.92 - (559.82 * math.cos(2 * lat)) + (1.175 * math.cos(4 * lat))
        
        return round(m_per_degree / self.ySize, 3)
    
    def shrinkGeoT(self, downsample_factor):
        """
        Function to modify gdal geo_transform to output at correct resolution for downsampled products.
        """
        
        geo_t = list(self.geo_t)
                
        geo_t[1] = 1. / int(round(self.xSize / float(downsample_factor)))
        geo_t[-1] = (1. / int(round(self.ySize / float(downsample_factor)))) * -1
        
        return tuple(geo_t)
    
    def __rebin(self):
        """
        Reduce array size to a new resolution by taking mean, to increase equivalent number of looks.
        # From https://stackoverflow.com/questions/8090229/resize-with-averaging-or-rebin-a-numpy-2d-array
        """
        
        # Update geo_t
        self.geo_t = self.shrinkGeoT(self.downsample_factor)
        
        # Take the mean of X/Y resolutions in metres.
        native_resolution = (self.xRes + self.yRes) / 2.

        # Update extents to new resolution
        new_xSize = int(round(self.xSize / float(self.downsample_factor)))
        new_ySize = int(round(self.ySize / float(self.downsample_factor)))
        
        self.xRes = (self.xSize * self.xRes) / new_xSize
        self.yRes = (self.ySize * self.yRes) / new_xSize
        self.xSize = new_xSize
        self.ySize = new_ySize
                
        # Update number of looks
        self.nLooks = self.nLooks * (self.downsample_factor ** 2)
    
    def __getMask(self, masked_px_count = False, output = False, show = False):
        """
        Loads the mask into a numpy array.
        """
        
        mask = biota.IO.loadArray(self.mask_path) != 255
                        
        # If resampling, this removes the mask from any pixel with >= 75 % data availability.
        if self.downsample_factor != 1 and not masked_px_count:
            mask = skimage.measure.block_reduce(mask, (self.downsample_factor, self.downsample_factor), np.sum) >= ((self.downsample_factor ** 2) * 0.25)
            
        # This is an option to return the sum of masked pixels. It's used to downsample the DN array.
        if self.downsample_factor != 1 and masked_px_count:
            mask = skimage.measure.block_reduce(mask, (self.downsample_factor, self.downsample_factor), np.sum)   
        
        if output: self.__outputGeoTiff(mask, 'Mask', dtype = gdal.GDT_Byte)
        
        if show: self.__showArray(mask, title = 'Mask', cmap = 'coolwarm')
        
        return mask
    
    def updateMask(self, filename, buffer_size = 0., classes = [], output = False, show = False):
        """
        Function to add further pixels to mask based on a numpy array, shapefile or GeoTiff file, with optional buffers.
        """
        
        # Load mask
        mask = biota.mask.updateMask(self, filename, buffer_size = buffer_size, classes = classes)
        
        # Add the new raster masks to the existing mask
        self.mask = np.logical_or(self.mask, mask)
    
        if output: self.__outputGeoTiff(self.mask, 'Mask', dtype = gdal.GDT_Byte)
        
        if show: self.__showArray(self.mask, title = 'Mask', cmap = 'coolwarm')
        
    def resetMask(self):
        """
        Function to reset a mask to the default.
        """
        
        self.mask = self.__getMask()
       
    def getDN(self, polarisation = 'HV', output = False, show = False):
        """
        Loads DN (raw) values into a numpy array.
        """
        
        assert polarisation == 'HH' or polarisation == 'HV', "polarisation must be either 'HH' or 'HV'."
               
        if polarisation == 'HV':
            DN = biota.IO.loadArray(self.HV_path)
        else:
            DN = biota.IO.loadArray(self.HH_path)
                
        # Rebin DN with mean. This is acceptable as the DN values are on a linear scale.
        if self.downsample_factor != 1:
            
            # Load the sum of DNs
            DN_sum = skimage.measure.block_reduce(DN, (self.downsample_factor, self.downsample_factor), np.sum)
            
            # Load the sum of contributing pixels
            block_sum = skimage.measure.block_reduce(np.ones_like(DN), (self.downsample_factor, self.downsample_factor), np.sum)
            
            # And the sum of masked pixels
            mask_sum = self.__getMask(masked_px_count = True)
            
            # Divide the sum of DNs by the sum of unmasked pixels to get the mean DN value
            DN = np.zeros_like(DN_sum)
            
            DN[self.mask == False] = (DN_sum.astype(np.float)[self.mask == False] / (block_sum - mask_sum)[self.mask == False]).astype(np.int)
        
        if output: self.__outputGeoTiff(DN, 'DN', dtype = gdal.GDT_Int32)
        
        if show: self.__showArray(DN, title = 'DN', cbartitle = 'Digital Number', cmap = 'Spectral_r')

        return np.ma.array(DN, mask = self.mask)

    def __getDay(self):
        '''
        '''
        
        day_after_launch = biota.IO.loadArray(self.date_path)
        
        # If required, downsample the dates array      
        if self.downsample_factor != 1:
            
            day_after_launch = skimage.measure.block_reduce(day_after_launch, (self.downsample_factor, self.downsample_factor), np.max)
        
        return day_after_launch
    
    def getDate(self, output = False, show = False):
        """
        Loads date values into a numpy array.
        """
        
        from osgeo import gdal

        # Don't rerun processing if already present in memory
        if not hasattr(self, 'date'):
            
            day_after_launch = self.__getDay()

            # Get list of unique dates in ALOS tile
            unique_days = np.unique(day_after_launch[day_after_launch != 0])
            
            # Days are counted from the launch date
            if self.satellite == 'ALOS-2':
                launch_date = dt.date(2014, 5, 24)
            else:
                launch_date = dt.date(2006, 1, 24)
            
            # Determine the Day of Year associated with each
            unique_dates = [launch_date  + dt.timedelta(days=int(d)) for d in unique_days]
            
            # Build a Day Of Year array
            dates = np.zeros_like(day_after_launch, dtype='datetime64[D]')
            dates_int = np.zeros_like(day_after_launch, dtype = np.int32)
            for day, date in zip(unique_days, unique_dates):
	        dates[day_after_launch == day] = np.datetime64(date,'D')
                dates_int[day_after_launch == day] = np.int(np.datetime64(date,'D').astype(dt.date).strftime('%Y%m%d'))
            
            # Save output to class
            self.date = dates
            self.date_int = dates_int
        
        if output: self.__outputGeoTiff(self.dates_int, 'Date', dtype = gdal.GDT_Int32)
        
        if show: print "Sorry, matplotlib doesn't support display of numpy datetime objects right now."
        
        return self.date
        
    def getDOY(self, output = False, show = False):
        """
        Loads day of year values into a numpy array.
        """

        from osgeo import gdal

        # Don't rerun processing if already present in memory
        if not hasattr(self, 'DOY'):
                        
            day_after_launch = self.__getDay()
            dates = self.getDate()       
            
            # Determine the Day of Year associated with each
            unique_dates = np.unique(dates)
            unique_doys = [(d.astype(dt.date) - dt.date(d.astype(dt.date).year,1,1)).days + 1 for d in unique_dates]
            unique_days = np.unique(day_after_launch)

            # Build a Day Of Year array
            DOY = np.zeros_like(dates, dtype = np.int16)
            for day, doy in zip(unique_days, unique_doys):
                DOY[day_after_launch == day] = doy
            
            # Save output to class
            self.DOY = DOY
            
        if output: self.__outputGeoTiff(self.DOY, 'DOY', dtype = gdal.GDT_Int32)
        
        if show: self.__showArray(self.DOY, title = 'DOY', cbartitle = 'Day', vmin = 0, vmax = 366, cmap = 'Spectral')
        
        return self.DOY       
    
    def getSM(self, output = False, show = False):
        """
        Loads a soil moisture map using the ESA CCI soil moisture product.
        """

        # Don't rerun processing if already present in memory
        if not hasattr(self, 'SM'):
            
            SM = biota.SM.getSM(self)
            
            # Keep masked values tidy
            SM.data[self.mask] = self.nodata
            
            # Save output to class
            self.SM = SM
        
        if output: self.__outputGeoTiff(self.SM, 'SM')

        if show: self.__showArray(self.SM, title = 'Soil Moisture', cbartitle = 'm^2/m^2', vmin = 0., vmax = 0.3, cmap = 'Blues')
        
        return self.SM
        
    def getGamma0(self, polarisation = 'HV', units = 'natural', output = False, show = False):
        """
        Calibrates data to gamma0 (baskscatter) in decibels or natural units.
        """
        
        assert units == 'natural' or units == 'decibels', "Units must be 'natural' or 'decibels'. You input %s."%units
        assert polarisation == 'HH' or polarisation == 'HV', "Polarisation must be 'HH' or 'HV'. You input %s."%polarisation

        # Calibrate DN to units of dB
        gamma0 = 10 * np.ma.log10(self.getDN(polarisation = polarisation).astype(np.float) ** 2) - 83. # units = decibels
        
        # Apply filter based on dB values
        if self.lee_filter:
            gamma0 = biota.filter.enhanced_lee_filter(gamma0, n_looks = self.nLooks)
        
        
        # Convert to natural units where specified
        if units == 'natural': gamma0 = 10 ** (gamma0 / 10.)
        
        # Keep masked values tidy
        gamma0.data[self.mask] = self.nodata
        gamma0.mask = self.mask
        
        if output: self.__outputGeoTiff(gamma0, 'Gamma0')
        
        if show:
            # Different display settings depending on options
            if polarisation == 'HH' and units == 'natural units': vmin, vmax = 0, 0.15
            if polarisation == 'HV' and units == 'natural units': vmin, vmax = 0, 0.06
            if polarisation == 'HH' and units == 'dB': vmin, vmax = -15, -5
            if polarisation == 'HV' and units == 'dB': vmin, vmax = -20, -10
            
            self.__showArray(gamma0, title = 'Gamma0 %s'%polarisation, cbartitle = units, vmin = vmin, vmax = vmax, cmap = 'Greys_r')

        return gamma0
    
    def getAGB(self, output = False, show = False):
        """
        Calibrates data to aboveground biomass (AGB).
        Placeholder equation to calibrate backscatter (gamma0) to AGB (tC/ha).
        """
        
        # Don't rerun processing if already present in memory
        if not hasattr(self, 'AGB'):
                    
            # ALOS-1
            if self.satellite == 'ALOS-1':
                AGB = 715.667 * self.getGamma0(units = 'natural', polarisation = 'HV') - 5.967
            
            # ALOS-2 (to calculate)
            elif self.satellite == 'ALOS-2':
                AGB = 715.667 * self.getGamma0(units = 'natural', polarisation = 'HV') - 5.967
                
            else:       
                raise ValueError("Unknown satellite named '%s'. self.satellite must be 'ALOS-1' or 'ALOS-2'."%self.satellite)
                        
            # Save output to class
            self.AGB = AGB
        
        # Keep masked values tidy
        #self.AGB.data[self.mask] = self.nodata
        self.AGB.mask = self.mask
        
        if output: self.__outputGeoTiff(self.AGB, 'AGB')
        
        if show: self.__showArray(self.AGB, title = 'AGB', cbartitle = 'tC/ha', vmin = 0, vmax = 40, cmap = 'YlGn')
        
        return self.AGB

    def getWoodyCover(self, output = False, show = False):
        """
        Get woody cover, based on a threshold of AGB.
        min_area in ha
        """
        
        from osgeo import gdal
        
        # Don't rerun processing if already present in memory
        if not hasattr(self, 'WoodyCover'):
            
            WoodyCover = self.getAGB() >= float(self.forest_threshold)
                
            if self.area_threshold > 0:
                
                # Calculate number of pixels in min_area (assuming input is given in hecatres)
                min_pixels = int(round(self.area_threshold / (self.yRes * self.xRes * 0.0001)))
                
                # Remove pixels that aren't part of a forest block of size at least min_pixels
                contiguous_area, _ = biota.indices.getContiguousAreas(WoodyCover, True, min_pixels = min_pixels)
                
                WoodyCover.data[contiguous_area == False] = False
                        
            # Save output to class
            self.WoodyCover = WoodyCover
        
        # Keep masked values tidy
        #self.WoodyCover.data[self.mask] = False
        self.WoodyCover.mask = self.mask
        
        if output: 
            
            # Convert data to integer type for output
            WoodyCover_out = self.WoodyCover.astype(np.uint8)
            
            self.__outputGeoTiff(WoodyCover_out, 'WoodyCover', dtype = gdal.GDT_Byte)
        
        if show: self.__showArray(self.WoodyCover, title = 'Woody Cover', cbartitle = '', vmin = 0, vmax = 1, cmap = 'summer_r')
        
        return self.WoodyCover
    
    def getForestPatches(self, output = False, show = False):
        """
        Get numbered forest patches, based on woody cover threshold.
        """
        
        from osgeo import gdal
        
        # Don't rerun processing if already present in memory
        if not hasattr(self, 'ForestPatches'):
            
            WoodyCover = self.getWoodyCover()
            
            # Calculate number of pixels in min_area
            min_pixels = int(round(self.area_threshold / (self.yRes * self.xRes * 0.0001)))
            
            # Get areas that meet that threshold
            _, ForestPatches = biota.indices.getContiguousAreas(WoodyCover, True, min_pixels = min_pixels)
            
            # Save output to class
            self.ForestPatches = ForestPatches
                  
        # Tidy up masked pixels      
        #ForestPatches.data[np.ma.getmaskarray(self.ForestPatches)] = self.nodata
        self.ForestPatches.mask = self.mask       
        
        if output: self.__outputGeoTiff(self.ForestPatches * 1, 'ForestPatches', dtype = gdal.GDT_Int32)
        
        if show: self.__showArray(self.ForestPatches, title = 'Forest patches', cbartitle = 'Patch ID', cmap = 'Spectral')
        
        return self.ForestPatches
    
    def __outputGeoTiff(self, data, output_name, dtype = 6):
        """
        Output a GeoTiff file.
        """
        
        from osgeo import gdal
        
        # Generate a standardised filename
        filename = self.output_pattern%output_name
        
        # Generate a nodata value appropriate for datatype
        nodata = self.__getNodata(dtype = dtype)
        
        # Write to disk
        biota.IO.outputGeoTiff(data, filename, self.geo_t, self.proj, output_dir = self.output_dir, dtype = dtype, nodata = nodata)
    
    def __showArray(self, data, title = '', cbartitle = '', vmin = None, vmax = None, cmap = None):
        """
        Display data from a tile.
        """
        
        biota.IO.showFigure(data, self.lat, self.lon, title = title, cbartitle = cbartitle, vmin = vmin, vmax = vmax, cmap = cmap)
        

class LoadChange(object):
    """
    Input is two mosaic tiles from LoadTile, will output maps and change statistics.
    """
        
    def __init__(self, tile_t1, tile_t2, change_intensity_threshold = 0.2, change_magnitude_threshold = 0., change_area_threshold = 0, output_dir = os.getcwd(), output = False):
        '''
        Initialise
        '''
        
        from osgeo import gdal
        
        self.tile_t1 = tile_t1
        self.tile_t2 = tile_t2
        
        # Ensure that tiles are compatable
        self.__testTiles()
        
        # Get basic properties
        self.lat = tile_t1.lat
        self.lon = tile_t1.lon
        self.xRes = tile_t1.xRes
        self.yRes = tile_t1.yRes
        self.xSize = tile_t1.xSize
        self.ySize = tile_t1.ySize
        
        self.hem_NS = tile_t1.hem_NS
        self.hem_EW = tile_t1.hem_EW
        
        # Get GDAL geotransform and projection
        self.geo_t = tile_t1.geo_t
        self.proj = tile_t1.proj
                
        # Change definitions
        self.change_intensity_threshold = change_intensity_threshold
        self.change_magnitude_threshold = change_magnitude_threshold
        self.change_area_threshold = change_area_threshold
        
        self.year_t1 = tile_t1.year
        self.year_t2 = tile_t2.year
        
        self.output_dir = output_dir
        self.output_pattern = self.__getOutputPattern()
        
        # Nodata currently hardwired to 99
        self.nodata = tile_t1.nodata
        self.nodata_byte = tile_t1.nodata_byte
        
        # Calculate combined mask
        self.mask = self.__combineMasks()
        
    def __testTiles(self):
        '''
        Test that input tiles are from reasonable lats/lons/years
        '''
        
        assert self.tile_t1.lat == self.tile_t2.lat and self.tile_t1.lon == self.tile_t2.lon, "Input tiles must be from the same location."
        assert self.tile_t1.year <= self.tile_t2.year, "Input tile_t2 must be from a later year than tile_t1."
        assert self.tile_t1.year != self.tile_t2.year, "Input tile_t1 cannot be from the same year as tile_t2."
        assert self.tile_t1.lee_filter == self.tile_t2.lee_filter, "Only one of the input tiles has been filtered. Both tiles should have the same pre-processing parameters."
        assert self.tile_t1.proj == self.tile_t2.proj, "Input tiles do not have the same projection."
        assert self.tile_t1.xSize == self.tile_t2.xSize and self.tile_t1.ySize == self.tile_t2.ySize, "Input tiles do not have the same resolution."
        assert self.tile_t1.geo_t == self.tile_t2.geo_t, "Input tiles do not have the same geo_transform."
        assert self.tile_t1.forest_threshold == self.tile_t2.forest_threshold, "'forest_threshold' must be identical for both input tiles."
        assert self.tile_t1.area_threshold == self.tile_t2.area_threshold, "'area_threshold' must be identical for both input tiles."
        assert self.tile_t1.downsample_factor == self.tile_t2.downsample_factor, "'downsample_facor' must be identical for both input tiles."


    def __combineMasks(self):
        '''
        Add together masks from tile_t1 and tile_t2
        '''
        
        return np.logical_or(self.tile_t1.mask, self.tile_t2.mask)
        
    def __getOutputPattern(self):
        """
        Generates a filename pattern for data output.
        """
        
        return '%s_%s_%s_%s%s.tif'%('%s', str(self.year_t1), str(self.year_t2), self.hem_NS + str(abs(self.lat)).zfill(2), self.hem_EW + str(abs(self.lon)).zfill(3))

    def __getNodata(self, dtype = 6):
        """
        Return nodata values
        """
        
        from osgeo import gdal
        
        # Generate a nodata value
        if dtype == gdal.GDT_Byte:
            nodata = 255
        else:
            nodata = 999999
        
        return nodata

    def updateMask(self, filename, buffer_size = 0., classes = [], output = False, show = False):
        """
        Function to add further pixels to mask based on a numpy array, shapefile or GeoTiff file, with optional buffers.
        """
        
        # Load mask
        mask = biota.mask.updateMask(self, filename, buffer_size = buffer_size, classes = classes)
        
        # Add the new raster masks to the existing mask
        self.mask = np.logical_or(self.mask, mask)
    
        if output: self.__outputGeoTiff(self.mask, 'Mask', dtype = gdal.GDT_Byte)
        
        if show: self.__showArray(self.mask, title = 'Mask', cmap = 'coolwarm')
        
    def resetMask(self):
        """
        Function to reset a mask to the default.
        """
        
        self.mask = self.__combineMasks()
    
    
    def getSMChange(self, output = False, show = False):
        """
        """
        
        # Only run processing if not already done
        if not hasattr(self, 'SMChange'):
            
            self.SMChange = self.tile_t2.getSM() - self.tile_t1.getSM()

        if output: self.__outputGeoTiff(self.SMChange, 'SMChange')
        
        if show: self.__showArray(self.SMChange, title = 'SM Change', cbartitle = 'm^2/m^2', vmin = -0.15, vmax = 0.15, cmap = 'RdBu')
        
        return self.SMChange
        
    def getAGBChange(self, output = False, show = False):
        '''
        '''
        
        # Only run processing if not already done
        if not hasattr(self, 'AGBChange'):
            
            self.AGB_change = self.tile_t2.getAGB() - self.tile_t1.getAGB()
        
        self.AGB_change.mask = self.mask     

        if output: self.__outputGeoTiff(self.AGB_change, 'AGBChange')
        
        if show: self.__showArray(self.AGB_change, title = 'AGB Change', cbartitle = 'tC/ha', vmin = -10, vmax = 10, cmap = 'RdBu')
        
        return self.AGB_change
        
        
    def getChangeType(self, output = False, show = False):
        '''
        Returns pixels that meet change detection thresholds for country.
        
        Args:
            forest_threshold = threshold above which a pixel is forest
            intensity_threshold = threshold of proportional change with which to accept a change as real
            area_threshold
        '''
        
        from osgeo import gdal
        
        # Only run processing if not already done
        if not hasattr(self, 'ChangeType'):

            # Pixels that move from forest to nonforest (F_NF) or vice versa (NF_F)
            F_NF = np.logical_and(self.tile_t1.getWoodyCover(), self.tile_t2.getWoodyCover() == False)
            NF_F = np.logical_and(self.tile_t1.getWoodyCover() == False, self.tile_t2.getWoodyCover())
            
            # Pixels that remain forest (F_F) or nonforest (NF_NF)
            F_F = np.logical_and(self.tile_t1.getWoodyCover(), self.tile_t2.getWoodyCover())
            NF_NF = np.logical_and(self.tile_t1.getWoodyCover() == False, self.tile_t2.getWoodyCover() == False)
            
            # Get pixels of change greater than intensity threshold
            CHANGE_INTENSITY = np.logical_or((self.getAGBChange() / self.tile_t1.getAGB()) >= self.change_intensity_threshold, (self.getAGBChange() / self.tile_t1.getAGB()) < (- self.change_intensity_threshold))
            
            # Get pixels of change greater than magnitude threshold
            CHANGE_MAGNITUDE = np.logical_or(self.getAGBChange() >= self.change_magnitude_threshold, self.getAGBChange() < (- self.change_magnitude_threshold))
            
            CHANGE = np.logical_and(CHANGE_INTENSITY, CHANGE_MAGNITUDE) 
            NOCHANGE = CHANGE == False
            
            # Trajectory (changes can be positive or negative)
            DECREASE = self.tile_t2.getAGB() < self.tile_t1.getAGB()
            INCREASE = self.tile_t2.getAGB() >= self.tile_t1.getAGB()
            
            # Get a minimum pixel extent. Loss/Gain events much have a spatial extent greater than min_pixels, and not occur in nonforest.
            if self.change_area_threshold > 0:
                
                min_pixels = int(round(self.change_area_threshold / (self.yRes * self.xRes * 0.0001)))
                
                # Get areas of change that meet minimum area requirement
                CHANGE_INCREASE, _ = biota.indices.getContiguousAreas(CHANGE & INCREASE & (NF_F | F_F), True, min_pixels = min_pixels)
                CHANGE_DECREASE, _ = biota.indices.getContiguousAreas(CHANGE & DECREASE & (F_NF | F_F), True, min_pixels = min_pixels)
                CHANGE = np.logical_or(CHANGE_INCREASE, CHANGE_DECREASE)
                NOCHANGE = CHANGE == False
            
            change_type = {}
            
            # These are all the possible definitions. Note: they *must* add up to one.
            change_type['deforestation'] = F_NF & CHANGE
            change_type['degradation'] = F_F & CHANGE & DECREASE
            change_type['minorloss'] = (F_F | F_NF) & NOCHANGE & DECREASE
            
            change_type['afforestation'] = NF_F & CHANGE
            change_type['growth'] = F_F & CHANGE & INCREASE
            change_type['minorgain'] = (F_F | NF_F) & NOCHANGE & INCREASE
                    
            change_type['nonforest'] = NF_NF
            
            # Also produce a c for output
            change_code = np.zeros((self.ySize, self.xSize), dtype = np.int8) + self.nodata_byte
            
            change_code[change_type['nonforest'].data] = 0
            change_code[change_type['deforestation'].data] = 1
            change_code[change_type['degradation'].data] = 2
            change_code[change_type['minorloss'].data] = 3
            change_code[change_type['minorgain'].data] = 4
            change_code[change_type['growth'].data] = 5
            change_code[change_type['afforestation'].data] = 6
                        
            # Save to class
            self.ChangeType = change_type
            self.ChangeCode = change_code
        
        self.ChangeType.mask = self.mask     
        self.ChangeCode.mask = self.mask    
        
        if output: self.__outputGeoTiff(self.ChangeCode, 'ChangeType', dtype = gdal.GDT_Byte)
        
        if show:
            # Hide minor gain, minor loss and nonforest in display output
            change_code_display = np.ma.array(self.ChangeCode, mask = np.zeros_like(self.ChangeCode, dtype = np.bool))
            change_code_display.mask[np.logical_or(np.logical_or(change_code_display == 3, change_code_display == 4), change_code_display == 0)] = True
            self.__showArray(change_code_display, title = 'Change type', cbartitle = 'Class', vmin = 1, vmax = 6, cmap = 'Spectral')
                
        return self.ChangeType
        
    
    def __sumChange(self, change_type, scale = 1):
        '''
        Function for scaling then summing and (optionally) scaling change statistics.
        '''
        
        return {k: np.sum(v * scale) for k, v in change_type.items()}
    
    def getAreaSum(self, proportion = False, output = False):
        '''
        Extract a change area in hectares. 
        '''
        
        # Classify change type
        change_type = self.getChangeType()
                        
        # Get area change in units of ha/pixel
        change_hectares = (self.mask == False) * (self.xRes * self.yRes * 0.0001)
        
        totals = self.__sumChange(change_type, scale = change_hectares)
        
        if proportion: 
            for change in totals:
                totals[change] = totals[change] / change_hectares.sum()
        
        if output: print 'TODO'
                
        return totals
    
    
    def getAGBSum(self, proportion = False, output = False):
        '''
        Extract a change magnitude in tonnes of carbon.
        TODO: Proportional measures need work.
        TODO: Add similar summary stats to LoadTile()?
        '''
        
        # Classify change type
        change_type = self.getChangeType()
        
        # Get AGB change in units of tC/pixel
        change_AGB = self.getAGBChange() * (self.xRes * self.yRes * 0.0001)
        
        totals = self.__sumChange(change_type, scale = change_AGB)
                
        if proportion: 
            for change_type in totals:    
                totals[change_type] = totals[change_type] / (self.tile_t1.getAGB() * self.xRes * self.yRes * 0.0001).sum()
        
        if output: print 'TODO'
                
        return totals
    
    
    def __outputGeoTiff(self, data, output_name, dtype = 6):
        """
        Output a GeoTiff file.
        """
                
        # Generate a standardised filename
        filename = self.output_pattern%output_name
        
        nodata = self.__getNodata(dtype = dtype)
        
        # Write to disk
        biota.IO.outputGeoTiff(data, filename, self.geo_t, self.proj, output_dir = self.output_dir, dtype = dtype, nodata = nodata)
    
    def __showArray(self, data, title = '', cbartitle = '', vmin = None, vmax = None, cmap = None):
        '''
        '''
        
        biota.IO.showFigure(data, self.lat, self.lon, title = title, cbartitle = cbartitle, vmin = vmin, vmax = vmax, cmap = cmap)

        



#!/usr/bin/env python

import matplotlib.pyplot as plt
import numpy as np
import scipy.ndimage as ndimage

import pdb

import biota.IO

def getContiguousAreas(data, value, min_pixels = 1):
    '''
    Get pixels that come from the same contigous area.
    
    Args:
        data: A numpy array
        value: Pixel value to include in contiguous_area (e.g. True for forest)
        min_area: What minimum area should be included (number of pixels, queens move)
    
    Returns:
        A binary array of pixels that meet the conditions
    '''
    
    from scipy.ndimage.measurements import label
    
    # If masked, we use this flag to save the mask for later.   
    masked = np.ma.isMaskedArray(data)
    
    # If any pixels are masked, we give them the value of the nearest valid pixel.
    if masked:
        mask = np.ma.getmaskarray(data)
        ind = ndimage.distance_transform_edt(mask, return_distances = False, return_indices = True)
        data = data.data[tuple(ind)]
    
    # Extract area that meets condition
    binary_array = (data == value) * 1
    
    # Label contigous areas with a number
    structure = ndimage.generate_binary_structure(2,2) # This connects diagonal elements
    location_id, n_areas = label(binary_array, structure = structure)
    
    # Get count of each value in array
    label_area = np.bincount(location_id.flatten())[1:]
    
    # Find those IDs that meet minimum area requirements
    include_id = np.arange(1, n_areas + 1)[label_area >= min_pixels]
    
    # Get a binary array of location_id pixels that meet the minimum area requirement
    contiguous_area = np.in1d(location_id, include_id).reshape(data.shape).astype(np.bool)
    
    # Return an array giving values to each area
    location_id[contiguous_area == False] = 0
    
    # Re-number location_ids 1 to n, given that some unique value shave now been removed
    location_id_unique, location_id_indices = np.unique(location_id, return_inverse = True)
    location_id = np.arange(0, location_id_unique.shape[0], 1)[location_id_indices].reshape(data.shape)
       
    # Put mask back in if input was a masked array
    if masked:
        contiguous_area = np.ma.array(contiguous_area, mask = mask)
        location_id = np.ma.array(location_id, mask = mask)

    return contiguous_area, location_id


def _calculatePatchSize(tile, patch_size):
    """
    Function to automatically calculate a patch_size
    
    Args:
        tile: Either an ALOS tile (biota.LoadTile()) or an ALOS change object (biota.LoadChange())
        patch_size = Number of pixels to build into a single patch. Set to 'auto' for an approx 100 x 100 output image
    
    Returns:
        The patch size
    """
    
    assert patch_size == 'auto' or type(patch_size) == float or type(patch_size) == int, "Patch size must be numeric or set to 'auto'."
    
    # Calculate patch_size (which aims for a 100 x 100 output image if set to auto.)
    if patch_size == 'auto': patch_size = int(round(((tile.xSize + tile.ySize) / 2.) / 100.))
    
    return patch_size


def _buildOutputArray(tile, patch_size, dtype = np.int):
    """
    Function to generate an appropriately sized output array
    
    Args:
        tile: Either an ALOS tile (biota.LoadTile()) or an ALOS change object (biota.LoadChange())
        patch_size = Number of pixels to build into a single patch. Set to 'auto' for an approx 100 x 100 output image
    
    Returns:
        An empty array
    """
    
    # Get nodata value
    if dtype == np.int:
        nodata = tile.nodata_byte
    else:
        nodata = tile.nodata
    
    # Calculate output output size based on reduction patch_size
    output_size = int(round(((tile.xSize + tile.ySize) / 2.) / patch_size))
    
    # Empty array for output
    output_array = np.ma.array(np.zeros((output_size, output_size), dtype = dtype), mask = np.zeros((output_size, output_size), dtype = np.bool)) + nodata
    
    return output_array


def _getBlocks(tile, array):
    """
    Calculate blocks based on an ALOS tile and a downsampled array.
    
    Args:
        tile: Either an ALOS tile (biota.LoadTile()) or an ALOS change object (biota.LoadChange())
        array: The downsampled numpy array (from _buildOutputArray()).
    
    Returns:
        A list of tuples with the y, x coordinates of the output image, and ymin, ymax, xmin, xmax of input image.
    """
    
    ns, ms, ymins, ymaxs, xmins, xmaxs = [], [], [], [], [], []
    
    for n, ymin in enumerate(np.linspace(0, tile.ySize, array.shape[0], endpoint = False)):
        ymax = ymin + (tile.ySize / array.shape[0])
        for m, xmin in enumerate(np.linspace(0, tile.xSize, array.shape[1], endpoint = False)):
            xmax = xmin + (tile.xSize / array.shape[1])
            
            ns.append(n); ms.append(m)
            ymins.append(int(round(ymin))); ymaxs.append(int(round(ymax)))
            xmins.append(int(round(xmin))); xmaxs.append(int(round(xmax)))
    
    return zip(ns, ms, ymins, ymaxs, xmins, xmaxs)  
    


def calculateLDI(tile, patch_size = 'auto', output = False, show = False):
    """
    Landcover Division Index (LDI) is an indicator of the degree of habitat coherence, ranging from 0 (fully contiguous) to 1 (very fragmented).
    
    LDI is defined as the probability that two randomly selected points in the landscape are situated in two different patches of the habitat (Jaeger 2000; Mcgarigal 2015).
    
    Note that fragmentation of both an undisturbed forest area and contiguous agriculture will both be low, so interpret this index carefully
        
    Args:
        tile: An ALOS tile from niota.LoadTile()
        patch_size: Number of pixels to build into a single patch. Set to 'auto' for an approx 100 x 100 output image
    
    Returns:
        An numpy array.
    """
    
    from osgeo import gdal
    
    def _computeLDI(unique_ids):
        '''
        Calculates an index of landscape heterogeneity
        '''
        
        # If masked array, separate data from mask
        if np.ma.isMaskedArray(unique_ids):
            mask = np.ma.getmask(unique_ids)
            unique_ids = np.ma.getdata(unique_ids)
        else:
            mask = np.zeros_like(unique_ids, dtype=np.bool)
                
        # Calculate LDI
        selection_1 = np.zeros_like(unique_ids,dtype = np.bool).ravel()
        selection_2 = np.zeros_like(unique_ids,dtype = np.bool).ravel()
        
        # Draw 1/25th of pixels at random
        selection_1[:(selection_1.shape[0]/5)] = True
        selection_2[:(selection_2.shape[0]/5)] = True
        
        # Shuffle to get random pixels
        np.random.shuffle(selection_1)
        np.random.shuffle(selection_2)
        
        selection_1 = selection_1.reshape(unique_ids.shape)
        selection_2 = selection_2.reshape(unique_ids.shape)
        
        # This ignores IDs where one or the other of selection falls in an area of nodata
        mask_selection = np.logical_or(mask[selection_1], mask[selection_2])
        ID1 = unique_ids[selection_1][mask_selection==False]
        ID2 = unique_ids[selection_2][mask_selection==False]
        
        # Shuffle again to randomise pairs spatially
        np.random.shuffle(ID1)
        np.random.shuffle(ID2)
        
        # Number of random points from different patches
        different_patches = np.sum(ID1 != ID2)
        
        # Divide by potential matches
        if (mask_selection==False).sum() > 0:
            LDI = int(round((float(different_patches) / (mask_selection==False).sum()) * 100))
            
        # Unless everything is masked, in which case return nan.
        else:
            LDI = np.nan
        
        return LDI
            
    # Create an output array
    patch_size = _calculatePatchSize(tile, patch_size)
    LDI = _buildOutputArray(tile, patch_size)
    
    # Load woody cover
    woody_cover = tile.getWoodyCover()       
            
    # For each patch...
    for n, m, ymin, ymax, xmin, xmax in _getBlocks(tile, LDI):
        
        # Extract the data
        this_patch = woody_cover[ymin:ymax, xmin:xmax]
            
        # Get connected patches of forest and nonforest
        _, forest_id = getContiguousAreas(this_patch, True)
        _, nonforest_id = getContiguousAreas(this_patch, False)
        
        # Supply a unique ID to each habitat patch, whether forest or nonforest.
        unique_ids = forest_id.copy()
        unique_ids[nonforest_id != 0] = forest_id[nonforest_id != 0] + (nonforest_id[nonforest_id != 0] + np.max(forest_id) + 1)
        
        #  If at least 50 % of data is present...
        if this_patch.mask.sum() <= ((patch_size ** 2) * 0.5):
            
            ## Compute LDI
            LDI.data[n, m] = _computeLDI(unique_ids)
            
        else:
            
            LDI.mask[n, m] = True
        
    # Output GeoTiff
    if output: biota.IO.outputGeoTiff(LDI, tile.output_pattern%'LDI', tile.shrinkGeoT(patch_size), tile.proj, output_dir = tile.output_dir, dtype = gdal.GDT_Int32, nodata = tile.nodata)
    
    # Display
    if show: biota.IO.showFigure(LDI, tile.lat, tile.lon, title = 'LDI', cbartitle = '%', vmin = 0, vmax = 100, cmap = 'Oranges')
            
    return LDI


def calculateLDIChange(tile_change, patch_size = 'auto', output = False, show = False):
    """
    Landcover Division Index (LDI) is an indicator of the degree of habitat coherence, ranging from 0 (fully contiguous) to 1 (very fragmented).
    
    This function calculates the change in LDI between two tiles.
    
        Args:
        tile_change: An ALOS change object from biota.LoadChange()
        patch_size: Number of pixels to build into a single patch. Set to 'auto' for an approx 100 x 100 output image
    
    Returns:
        An numpy array.
        
    """
    
    from osgeo import gdal
    
    patch_size = _calculatePatchSize(tile_change, patch_size)
    
    LDI_t1 = calculateLDI(tile_change.tile_t1, patch_size = patch_size)
    LDI_t2 = calculateLDI(tile_change.tile_t2, patch_size = patch_size)
    
    LDI_change = LDI_t2 - LDI_t1
    
    # Output GeoTiff
    if output: biota.IO.outputGeoTiff(LDI_change, tile_change.output_pattern%'LDIChange', tile_change.shrinkGeoT(patch_size), tile_change.proj, output_dir = tile_change.output_dir, dtype = gdal.GDT_Int32, nodata = tile_change.nodata)
    
    # Display
    if show: biota.IO.showFigure(LDI_change, tile_change.lat, tile_change.lon, title = 'LDI Change', cbartitle = '%', vmin = -25, vmax = 25, cmap = 'RdBu_r')
    
    return LDI_change
            
            

def calculateTWC(tile, patch_size = 'auto', output = False, show = False):
    """
    Total woody cover (TWC) describes the proportion of woody cover in a downsampled image.
    
    Args:
        tile: An ALOS tile from biota.LoadTile()
        patch_size: Number of pixels to build into a single patch. Set to 'auto' for an approx 100 x 100 output image
    
    Returns:
        A numpy array,
    """

    from osgeo import gdal

    # Create an output array
    patch_size = _calculatePatchSize(tile, patch_size)
    TWC = _buildOutputArray(tile, patch_size)

    # Load woody cover
    woody_cover = tile.getWoodyCover()
    
    # For each patch...
    for n, m, ymin, ymax, xmin, xmax in _getBlocks(tile, TWC):
        
        # Extract the data
        WC = woody_cover[ymin:ymax, xmin:xmax]
         
        #  If at least 50 % of data is present...
        if WC.mask.sum() <= ((patch_size ** 2) * 0.5):
            
            # Calculate proportion of woody cover in patch
            TWC.data[n, m] = int(round((float(WC.sum()) / ((patch_size ** 2) - WC.mask.sum())) * 100))
            
        else:
            
            TWC.mask[n, m] = True
   
    # Output GeoTiff
    if output: biota.IO.outputGeoTiff(TWC, tile.output_pattern%'TWC', tile.shrinkGeoT(patch_size), tile.proj, output_dir = tile.output_dir, dtype = gdal.GDT_Int32, nodata = tile.nodata)
        
    # Display
    if show: biota.IO.showFigure(TWC, tile.lat, tile.lon, title = 'TWC', cbartitle = '%', vmin = 0, vmax = 100, cmap = 'YlGn')
        
    return TWC


def calculateWCC(tile_change, patch_size = 'auto', output = False, show = False):
    """
    Woody cover change (WCC) describes the loss of woody cover between two images.
    
    Args:
        tile_change: An ALOS change object from biota.LoadChange()
        patch_size: Number of pixels to build into a single patch. Set to 'auto' for an approx 100 x 100 output image
    
    Returns:
        A numpy array.    
    """
    
    from osgeo import gdal
    
    # Create an output array
    patch_size = _calculatePatchSize(tile_change, patch_size)
    WCC = _buildOutputArray(tile_change, patch_size)
    
    # Get total woody cover for time 1 and time 2
    TWC_t1 = calculateTWC(tile_change.tile_t1, patch_size = patch_size)
    TWC_t2 = calculateTWC(tile_change.tile_t2, patch_size = patch_size)
        
    # Woody cover change is defined as the difference between the two
    WCC = TWC_t2 - TWC_t1

    # Output GeoTiff
    if output: biota.IO.outputGeoTiff(WCC, tile_change.output_pattern%'WCC', tile_change.shrinkGeoT(patch_size), tile_change.proj, output_dir = tile_change.output_dir, dtype = gdal.GDT_Int32, nodata = tile_change.nodata_byte)
    
    # Display
    if show: biota.IO.showFigure(WCC, tile_change.lat, tile_change.lon, title = 'WCC', cbartitle = '%', vmin = -30, vmax = 30, cmap = 'RdBu')


def calculateProportionalChange(tile_change, change_type, patch_size = 'auto', output = False, show = False):
    """
    Calculate the proportion of a downsampled regions subject to a class of change.
    
    Args:
        tile_change: An ALOS change object from biota.LoadChange()
        change_type: String representing a change from tile_change.getChangeType(). e.e. 'deforestation' or 'degradation'.
        patch_size: Number of pixels to build into a single patch. Set to 'auto' for an approx 100 x 100 output image
    
    Returns:
        a numpy array
    """
    
    from osgeo import gdal
    
     # Create an output array
    patch_size = _calculatePatchSize(tile_change, patch_size)
    change_downsampled = _buildOutputArray(tile_change, patch_size)
    
    change = tile_change.getChangeType()[change_type]
    
    # For each patch...
    for n, m, ymin, ymax, xmin, xmax in _getBlocks(tile_change, change_downsampled):
        
        # Extract the data
        C = change[ymin:ymax, xmin:xmax]
        
        #  If at least 50 % of data is present...
        if C.mask.sum() <= ((patch_size ** 2) * 0.5):
            
            # Calculate proportion of patch subject to change_type
            change_downsampled.data[n, m] = int(round((float(C.sum()) / ((patch_size ** 2) - C.mask.sum())) * 100))

        else:
            change_downsampled.mask[n, m] = True

    # Output GeoTiff
    if output: biota.IO.outputGeoTiff(change_downsampled, tile_change.output_pattern%'%sDownsampled'%change_type.title(), tile_change.shrinkGeoT(patch_size), tile_change.proj, output_dir = tile_change.output_dir, dtype = gdal.GDT_Int32, nodata = tile_change.nodata)
    
    # Display
    if show: biota.IO.showFigure(change_downsampled, tile_change.lat, tile_change.lon, title = '%s Downsampled'%change_type.title(), cbartitle = '%', vmin = 0, vmax = 50, cmap = 'Spectral_r')

    
    return change_downsampled



#!/usr/bin/env python


# This is a set of scripts for calibration of a biomass-backscatter curve

import numpy as np
import scipy.stats

import biota
import biota.mask

import pdb



def extractGamma0(shp, plot_field, agb_field, tile_example, verbose = False, units = 'natural'):
    '''
    Extract gamma0 from ALOS tiles.
    
    Args:
        shp: A shapefile containing plot data
        plot_field: The shapefile field containing plot names
        agb_field: The shapefile field containing AGB estimates
        tile_example: An exemplar ALOS tile (from LoadTile()) which contains information on the processing chain.
        #TODO: year?
    Returns:
        A dictionary containing plot names, AGB and gamma0 values
    '''
    
    # Use example data to get processing steps
    downsample_factor = tile_example.downsample_factor
    lee_filter = tile_example.lee_filter
    year = tile_example.year
    dataloc = tile_example.dataloc
    
    # Extract relevant info from shapefile
    plot_names = biota.mask.getField(shp, plot_field) 
    agb = biota.mask.getField(shp, agb_field).astype(np.float)
    
    # Identify tiles that contain gamma0 data for the shapefile
    tiles = biota.mask.getTilesInShapefile(shp)
    
    # Check whether all the necessary tiles are all present. Decide what to do if they aren't.
    # TODO
    
    # Generate output array for backscatter
    gamma0_hv = np.empty_like(agb, dtype = np.float32)
    gamma0_hv[:] = np.nan
    
    gamma0_hh = gamma0_hv.copy()
    doy = gamma0_hv.copy()
    
    # For each tile covered by shapefile
    for lat, lon in tiles:
        
        if verbose: print 'Doing lat: %s, lon: %s'%(str(lat), str(lon))
        
        # Load tile
        tile = biota.LoadTile(dataloc, lat, lon, year, downsample_factor = downsample_factor, lee_filter = lee_filter)
        
        # Get backscatter (both polarisations) and DOY
        data_gamma0_hv = tile.getGamma0(polarisation = 'HV', units = units)
        data_gamma0_hh = tile.getGamma0(polarisation = 'HH', units = units)
        data_doy = tile.getDOY()
        
        # Mask out each plot
        plot_mask = biota.mask.rasterizeShapefile(tile, shp, location_id = True, buffer_size = 0.)
        
        # Extract values for each plot
        for n in np.unique(plot_mask[plot_mask != 0]):
            
            # Get mask for plot and tile
            this_mask = np.logical_and(plot_mask == n, tile.mask == False)
            
            # Add metrics to output array
            gamma0_hv[n-1] = np.mean(data_gamma0_hv[this_mask])
            gamma0_hh[n-1] = np.mean(data_gamma0_hh[this_mask])
            doy[n-1] = np.median(data_doy[this_mask])
    
    # Return data as a dictionary
    data_dict = {}
    data_dict['plot_name'] = plot_names
    data_dict['AGB'] = agb
    data_dict['gamma0_HH'] = gamma0_hh
    data_dict['gamma0_HV'] = gamma0_hv
    data_dict['DOY'] = doy
        
    return data_dict


def fitLinearModel(data_dict):
    '''
    Fit a linear model to relate AGB to gamma0 backscatter.
    
    Args:
        data_dict: Dictionary output from extractGamma0() function.
    Returns:
        model slope, model intercept
    '''
    
    # Select only data that have values. NaN can arise in cases of masked areas in ALOS tiles
    sel = np.logical_and(np.isfinite(data_dict['gamma0_HV']), np.isfinite(data_dict['AGB']))
    
    assert sel.sum() > 0, "No usable data in data_dict."
    
    slope, intercept, r_value, p_value, stderr = scipy.stats.linregress(data_dict['gamma0_HV'][sel], data_dict['AGB'][sel])
    
    print "r-squared:", r_value**2
    print "p value:", p_value
    
    return slope, intercept



import itertools
import math
import numpy as np
import os
from PIL import Image, ImageDraw
import scipy.ndimage as ndimage

import biota.IO

import pdb


def dilateMask(mask, buffer_px):
    """
    Dilate a boolean (True/False) numpy array by a specified number of pixels.
        
    Args:
        mask: A boolean (True/False) numpy array, with 'True' representing locations to add a buffer.
        buffer_px: A number of pixels to add around each 'True' array element.

    Returns:
        The mask array with dilated 'True' locations.
    """
        
    mask_dilated = ndimage.morphology.binary_dilation(mask, iterations = buffer_px)
    
    return mask_dilated


def _coordinateTransformer(shp):
    """
    Generates function to transform coordinates from a source shapefile CRS to EPSG.
    
    Args:
        shp: Path to a shapefile.
    
    Returns:
        A function that transforms shapefile points to EPSG.
    """
    
    from osgeo import ogr, osr
        
    driver = ogr.GetDriverByName('ESRI Shapefile')
    ds = driver.Open(shp)
    layer = ds.GetLayer()
    spatialRef = layer.GetSpatialRef()
    
    # Create coordinate transformation
    inSpatialRef = osr.SpatialReference()
    inSpatialRef.ImportFromWkt(spatialRef.ExportToWkt())

    outSpatialRef = osr.SpatialReference()
    outSpatialRef.ImportFromEPSG(4326)

    coordTransform = osr.CoordinateTransformation(inSpatialRef, outSpatialRef)
    
    return coordTransform


def _world2Pixel(geo_t, x, y, buffer_size = 0):
    """
    Uses a gdal geomatrix (ds.GetGeoTransform()) to calculate the pixel location of a geospatial coordinate.
    Modified from: http://geospatialpython.com/2011/02/clip-raster-using-shapefile.html.
    
    Args:
        geo_t: A gdal geoMatrix (ds.GetGeoTransform().
        x: x coordinate in map units.
        y: y coordinate in map units.
        buffer_size: Optionally specify a buffer size. This is used when a buffer has been applied to extend all edges of an image, as in rasterizeShapfile().
    
    Returns:
        A tuple with pixel/line locations for each input coordinate.
    """
    ulX = geo_t[0] - buffer_size
    ulY = geo_t[3] + buffer_size
    xDist = geo_t[1]
    yDist = geo_t[5]
    
    pixel = int((x - ulX) / xDist)
    line = int((y - ulY) / yDist)
    
    return (pixel, line)


def getField(shp, field):
    '''
    Get values from a field in a shapefile attribute table.
    
    Args:
        shp: A string pointing to a shapefile
        field: A string with the field name of the attribute of interest
    
    Retuns:
        An array containing all the values of the specified attribute
    '''
    
    import shapefile
    
    assert os.path.isfile(shp), "Shapefile %s does not exist."%shp
    
    # Read shapefile
    sf = shapefile.Reader(shp)
    
    # Get the column number of the field of interest
    for n, this_field in enumerate(sf.fields[1:]):
        
        fieldname = this_field[0]
        
        if fieldname == field:
            
            field_n = n
    
    assert 'field_n' in locals(), "Attribute %s not found in shapefile."%str(field)
    
    # Extract data type from shapefile. Interprets N (int), F (float) and C (string), sets others to string.
    this_dtype = sf.fields[1:][field_n][1]
    
    if this_dtype == 'N':
        dtype = np.int
    elif this_dtype == 'F':
        dtype = np.float32
    elif this_dtype == 'C':
        dtype = np.str
    else:
        dtype = np.str
    
    value_out = []
    
    # Cycle through records:
    for s in sf.records():
        value_out.append(s[field_n])
        
    return np.array(value_out, dtype = dtype)


def getBBox(shp, field, value):
    '''
    Get the bounding box of a shape in a shapefile.
    
    Args:
        shp: A string pointing to a shapefile
        field: A string with the field name of the attribute of interest
        value: The value of the field for the shape of interest
    
    Retuns:
        An list with the bounding box in the format [minlon, minlat, maxlon, maxlat]
    '''
        
    import shapefile
    
    assert os.path.isfile(shp), "Shapefile %s does not exist."%shp

    assert (np.sum(getField(shp, field) == value) > 1) == False, "The value name in a field must be unique. In the field %s there are %s records with value %s."%(str(field), str(np.sum(getField(shp, field) == value)), str(value))
    
    # Read shapefile
    sf = shapefile.Reader(shp)
    
    shapes = np.array(sf.shapes())
    
    # Get bounding box    
    bbox = shapes[getField(shp, field) == value][0].bbox
    
    return bbox

def maskArray(tile, array, classes = [], buffer_size = 0.):
    '''
    Extract a mask from a numpy array based on specified classes

    Args:
        tile: An ALOS tile (biota.LoadTile())
        array: A numpy array
        classes: A list of values to add to be masked
        buffer_size: Optionally specify a buffer to add around maked pixels, in meters.
    
    Returns:
        A numpy array with a boolean mask
    '''
    
    assert (tile.ySize, tile.xSize) == array.shape, "Numpy array shape must be identical to the tile extent."
    
    # If a binary mask is input, assume that True should be added to mask
    if array.dtype == np.bool and classes == []:
        classes = [True]
    
    # Identify pixels that are classes to be masked
    mask = np.in1d(array, classes).reshape(array.shape)
            
    if buffer_size > 0.:
        
        # Determine the size of the buffer in pixels
        buffer_px = int(buffer_size / ((tile.xRes + tile.yRes) / 2.))
        
        # Dilate the mask
        mask = dilateMask(mask, buffer_px)
        
    return mask


def maskRaster(tile, raster, classes = [], buffer_size = 0.):
    '''
    Extract a mask from a GeoTiff based on specified classes
    
    Args:
        tile: An ALOS tile (biota.LoadTile())
        raster: A GeoTiff or VRT file, with integer values
        classes: A list of values to add to be masked
        buffer_size: Optionally specify a buffer to add around maked pixels, in meters.
    
    Returns:
        A numpy array with a boolean mask
    '''
    
    from osgeo import gdal
    
    assert raster.rstrip('/').split('.')[-1] == 'tif' or raster.rstrip('/').split('.')[-1] == 'tiff' or raster.rstrip('/').split('.')[-1] == 'vrt', "raster input must be a GeoTiff or VRT file."
    
    raster = os.path.expanduser(raster)
    assert os.path.exists(raster), "GeoTiff file %s does not exist in the file system."%raster
    
    # Load raster + reporject to match tile
    reampled_image = biota.IO.loadRaster(raster, tile)
    
    # Identify pixels that are classes to be masked
    mask = maskArray(tile, reampled_image, classes = classes, buffer_size = buffer_size)
    
    return mask


def maskShapefile(tile, shp, buffer_size = 0., field = None, value = None, location_id = False):
    """
    Rasterize points, lines or polygons from a shapefile to match ALOS mosaic data.
        
    Args:
        tile: An ALOS tile (biota.LoadTile())
        shp: Path to a shapefile consisting of points, lines and/or polygons. This does not have to be in the same projection as ds
        buffer_size: Optionally specify a buffer to add around features of the shapefile, in meters.
        field: Optionally specify a single shapefile field to extract (you must also specify its value)
        value: Optionally specify a single shapefile field value to extract (you must also specify the field name)
        location_id: Set True to return a unique ID for each masked shape. Note: This is not zero indexed, but starts at 1.

    Returns:
        A numpy array with a boolean (or integer) mask delineating locations inside and outside the shapefile and optional buffer.
    """
    
    import shapefile
    from osgeo import gdalnumeric
    
    assert np.logical_or(np.logical_and(field == None, value == None), np.logical_and(field != None, value != None)), "If specifying field or value, both must be defined. At present, field = %s and value = %s"%(str(field), str(value))
    
    shp = os.path.expanduser(shp)
    assert os.path.exists(shp), "Shapefile %s does not exist in the file system."%shp
    
    # Determine the size of the buffer in degrees
    buffer_size_degrees = buffer_size / (((tile.xRes * tile.xSize) + (tile.yRes * tile.ySize)) / 2.)
    
    # Determine size of buffer to place around lines/polygons
    buffer_px = int(round(buffer_size_degrees / tile.geo_t[1]))
    
    # Create output image. Add a buffer around the image array equal to the maxiumum dilation size. This means that features just outside ALOS tile extent can contribute to dilated mask.
    rasterPoly = Image.new("I", (tile.xSize + (buffer_px * 2), tile.ySize + (buffer_px * 2)), 0)
    rasterize = ImageDraw.Draw(rasterPoly)
    
    # The shapefile may not have the same CRS as ALOS mosaic data, so this will generate a function to reproject points.
    coordTransform = _coordinateTransformer(shp)
    
    # Read shapefile
    sf = shapefile.Reader(shp) 
    
    # Get shapes
    shapes = np.array(sf.shapes())
    
    # If extracting a mask for just a single field.
    if field != None:
        
        shapes = shapes[getField(shp, field) == value]
            
    # For each shape in shapefile...
    for n, shape in enumerate(shapes):
                
        # Get shape bounding box
        sxmin, symin, sxmax, symax = shape.bbox
        
        # Transform bounding box points
        sxmin, symin, z = coordTransform.TransformPoint(sxmin, symin)
        sxmax, symax, z = coordTransform.TransformPoint(sxmax, symax)
        
        # Go to the next record if out of bounds
        if sxmax < tile.geo_t[0] - buffer_size_degrees: continue
        if sxmin > tile.geo_t[0] + (tile.geo_t[1] * tile.xSize) + buffer_size_degrees: continue
        if symax < tile.geo_t[3] + (tile.geo_t[5] * tile.ySize) + buffer_size_degrees: continue
        if symin > tile.geo_t[3] - buffer_size_degrees: continue
        
        #Separate polygons with list indices
        n_parts = len(shape.parts) #Number of parts
        indices = shape.parts #Get indices of shapefile part starts
        indices.append(len(shape.points)) #Add index of final vertex
        
        for part in range(n_parts):
            
            start_index = shape.parts[part]
            end_index = shape.parts[part+1]
            
            points = shape.points[start_index:end_index] #Map coordinates
            pixels = [] #Pixel coordinantes
            
            # Transform coordinates to pixel values
            for p in points:
                
                # First update points from shapefile projection to ALOS mosaic projection
                lon, lat, z = coordTransform.TransformPoint(p[0], p[1])

                # Then convert map to pixel coordinates using geo transform
                pixels.append(_world2Pixel(tile.geo_t, lon, lat, buffer_size = buffer_size_degrees))

            # Draw the mask for this shape...
            # if a point...
            if shape.shapeType == 0:
                rasterize.point(pixels, n+1)

            # a line...
            elif shape.shapeType == 3:
                rasterize.line(pixels, n+1)
  
            # or a polygon.
            elif shape.shapeType == 5:  
                rasterize.polygon(pixels, n+1)
        
    #Converts a Python Imaging Library array to a gdalnumeric image.
    mask = gdalnumeric.fromstring(rasterPoly.tobytes(),dtype=np.uint32)
    mask.shape = rasterPoly.im.size[1], rasterPoly.im.size[0]
    
    # If any buffer pixels are slected, dilate the masked area by buffer_px pixels
    if buffer_px > 0:
        mask = dilateMask(mask, buffer_px)
    
    # Get rid of image buffer
    mask = mask[buffer_px:mask.shape[0]-buffer_px, buffer_px:mask.shape[1]-buffer_px]
    
    if location_id == False:
        # Get rid of record numbers
        mask = mask > 0
    
    return mask


def getTilesInShapefile(shp, field = None, value = None):
    """
    Identify all the ALOS tiles that fall within a shapefile.
    
    Args:
        shp: Path to a shapefile consisting of polygons. This can be in any projection.
    
    Returns:
        The lat/lon indicators of which ALOS tiles are covered by the shapefile
    """

    import shapefile

    assert np.logical_or(np.logical_and(field == None, value == None), np.logical_and(field != None, value != None)), "If specifying field or value, both must be defined. At present, field = %s and value = %s"%(str(field), str(value))    
    
    # The shapefile may not have the same CRS as ALOS mosaic data, so this will generate a function to reproject points.    
    coordTransform = _coordinateTransformer(shp)
    
    lats, lons = [], []
    tiles_to_include = set([])
    
    # Get shapes
    shapes = np.array(shapefile.Reader(shp).shapes())
    
    if field != None:    
        shapes = shapes[getField(shp, field) == value]
        
    for shape in shapes:
        
        # Get the bbox for each shape in the shapefile
        lonmin, latmin, lonmax, latmax = shape.bbox
        
        # Transform points to WGS84
        lonmin, latmin, z = coordTransform.TransformPoint(lonmin, latmin)
        lonmax, latmax, z = coordTransform.TransformPoint(lonmax, latmax)
        
        # Get the tiles that cover the area of the shapefile
        latrange = range(int(math.ceil(latmin)), int(math.ceil(latmax)+1), 1)
        lonrange = range(int(math.floor(lonmin)), int(math.floor(lonmax)+1), 1)
        tiles = list(itertools.product(latrange,lonrange))
        
        # Add them to tiles_to_include if not already tere
        [tiles_to_include.add(t) for t in tiles]
    
    return sorted(list(tiles_to_include))


def updateMask(tile, filename, buffer_size = 0., classes = []):
    """
    Function to generate a VRT, GeoTiff, shapefile, or numpy array mask to match an ALOS tile.
    
    Args:
        ::
        tile: An ALOS tile (biota.LoadTile())
        filename: A GeoTiff, VRT or shapefile (string)m or a numpy array
        buffer_size: Optionally specify a buffer to add around maked pixels, in meters.
        classes: A list of values to add to be masked from a raster input. May be omitted where a boolean numpy array is input.
    
    Returns:
        A boolean numpy array
    """
    
    if type(filename) == str:
        
        file_type = filename.split('/')[-1].split('.')[-1]   
        
        assert file_type in ['shp', 'tif', 'tiff', 'vrt'], "Input must be a numpy array, GeoTiff, VRT, or a shapefile."
        
    elif type(filename) == np.ndarray or type(filename) == np.ma.core.MaskedArray:
        file_type = 'array'
        
    else:
        
        assert False, "Input must be a numpy array, GeoTiff, VRT, or a shapefile."
            
    if file_type == 'shp':
    
        # Rasterize the shapefile, optionally with a buffer
        mask = biota.mask.maskShapefile(tile, filename, buffer_size = buffer_size)
    
    elif file_type in ['tif', 'tiff', 'vrt']:
        
        assert classes != [], "If adding a GeoTiff or VRT file to the mask, you must also specify the class values to add to the mask (e.g. classes = [20, 160, 170, 190, 210])."
        
        # Resample and extract values from shapefile, optionally with a buffer
        mask = biota.mask.maskRaster(tile, filename, classes = classes, buffer_size = buffer_size)
    
    else:
        
        if filename.dtype != np.bool:
            assert classes != [], "If adding a non-boolean numpy array file to the mask, you must also specify the class values to add to the mask (e.g. classes = [20, 160, 170, 190, 210])."
        
        mask = biota.mask.maskArray(tile, filename, classes = classes, buffer_size = buffer_size)
    
    return mask


#!/usr/bin/env python

import numpy as np
import signal
import scipy.ndimage as ndimage

import pdb

"""
This file contains scripts to filter ALOS data.
"""

def enhanced_lee_filter(img, window_size = 7, n_looks = 16):
    '''
    Filters a masked array with the enhanced lee filter.
    
    Based on formulation here: http://www.pcigeomatics.com/geomatica-help/concepts/orthoengine_c/Chapter_825.html
    
    Args:
        img: A masked array
    Returns:
        A masked array with a filtered verison of img
    '''
    
    assert type(window_size == int), "Window size must be an integer. You input the value: %s"%str(window_size)
    assert (window_size % 2) == 1, "Window size must be an odd number. You input the value: %s"%str(window_size)
    assert window_size > 3, "Window size must be at least 3. You input the value: %s"%str(window_size)

    # Inner function to calculate mean with a moving window and a masked array
    def _window_mean(img, window_size = 3):
        '''
        Based on https://stackoverflow.com/questions/18419871/improving-code-efficiency-standard-deviation-on-sliding-windows
        '''
        
        from scipy import signal
        
        c1 = signal.convolve2d(img, np.ones((window_size, window_size)) / (window_size ** 2), boundary = 'symm')
        
        border = window_size/2
        
        return c1[border:-border, border:-border]
        
    # Inner function to calculate standard deviation with a moving window and a masked array
    def _window_stdev(img, window_size = 3):
        '''
        Based on https://stackoverflow.com/questions/18419871/improving-code-efficiency-standard-deviation-on-sliding-windows
        and http://nickc1.github.io/python,/matlab/2016/05/17/Standard-Deviation-(Filters)-in-Matlab-and-Python.html
        '''
        
        from scipy import signal
        
        c1 = signal.convolve2d(img, np.ones((window_size, window_size)) / (window_size ** 2), boundary = 'symm')
        c2 = signal.convolve2d(img*img, np.ones((window_size, window_size)) / (window_size ** 2), boundary = 'symm')
        
        border = window_size / 2
        
        variance = c2 - c1 * c1
        variance[variance < 0] += 0.01 # Prevents divide by zero errors.
        
        return np.sqrt(variance)[border:-border, border:-border]
    
    # Damping factor, set to 1 which is adequate for most SAR images
    k = 1
        
    cu = (1./n_looks) ** 0.5
    cmax =  (1 + (2./n_looks)) ** 0.5
    
    
    # Interpolate across nodata areas. No standard Python filters understand nodata values; this is a simplification
    
    indices = ndimage.distance_transform_edt(img.mask, return_distances = False, return_indices = True)
    data = img.data[tuple(indices)]
    
    img_mean = _window_mean(data, window_size = window_size)
    img_std = _window_stdev(data, window_size = window_size)

    ci = img_std / img_mean
    ci[np.isfinite(ci) == False] = 0.
    
    W = np.zeros_like(ci)
    
    # There are three conditions in the enhanced lee filter
    W[ci <= cu] = 1.
    W[ci >= cmax] = 0.
    
    s = np.logical_and(ci > cu, ci < cmax)
    W[s] = np.exp((-k * (ci[s] - cu)) / (cmax - ci[s]))
        
    img_filtered = (img_mean * W) + (data * (1. - W))
    
    img_filtered = np.ma.array(img_filtered, mask = np.logical_or(np.isnan(img_filtered), img.mask))
    
    img_filtered.data[img_filtered.mask] = 0.
    
    return img_filtered



# Toolset to calibrate change attribution algorithm

import math
import numpy as np
import scipy.stats as stats
import scipy.ndimage
from scipy.ndimage.measurements import label

import biota.indices

import matplotlib.pyplot as plt
import pdb



def getChange(data_t1, data_t2, forest_threshold = 10., intensity_threshold = 0.2, area_threshold = 0, output = False):
    """
    Returns pixels that meet change detection thresholds for country.
    
    Args:
        forest_threshold = threshold above which a pixel is forest
        intensity_threshold = threshold of proportional change with which to accept a change as real
        area_threshold
    """
    
    from osgeo import gdal
    
    AGB_t1 = data_t1.getAGB()
    AGB_t2 = data_t2.getAGB()
    
    # Combine masks, for case where they differ. TODO: Set alternative where AGB is not masked by user.
    mask = np.logical_or(AGB_t1.mask, AGB_t2.mask)
    AGB_t1.mask = mask
    AGB_t2.mask = mask
    
    # Pixels that move from forest to nonforest (F_NF) or vice versa (NF_F)
    F_NF = np.logical_and(AGB_t1 >= forest_threshold, AGB_t2 < forest_threshold)
    NF_F = np.logical_and(AGB_t1 < forest_threshold, AGB_t2 >= forest_threshold)
    
    # Pixels that remain forest (F_F) or nonforest (NF_NF)
    F_F = np.logical_and(AGB_t1 >= forest_threshold, AGB_t2 >= forest_threshold)
    NF_NF = np.logical_and(AGB_t1 < forest_threshold, AGB_t2 < forest_threshold)
    
    # Get change pixels given requirements
    CHANGE = np.logical_or(((AGB_t1 - AGB_t2) / AGB_t1) >= intensity_threshold, ((AGB_t1 - AGB_t2) / AGB_t1) < (- intensity_threshold))
    NOCHANGE = CHANGE == False
    
    # Trajectory (changes can be positive or negative)
    DECREASE = AGB_t2 < AGB_t1
    INCREASE = AGB_t2 >= AGB_t1
    
    # Get a minimum pixel extent. Loss/Gain events much have a spatial extent greater than min_pixels, and not occur in nonforest.
    if area_threshold > 0:
        
        pixel_area = data_t1.xRes * data_t1.yRes * 0.0001
        min_pixels = int(round(area_threshold / pixel_area))
        
        # Get areas of change that meet minimum area requirement
        CHANGE_INCREASE, _ = biota.indices.getContiguousAreas(CHANGE & INCREASE & (NF_F | F_F), True, min_pixels = min_pixels)
        CHANGE_DECREASE, _ = biota.indices.getContiguousAreas(CHANGE & DECREASE & (F_NF | F_F), True, min_pixels = min_pixels)
        CHANGE = np.logical_or(CHANGE_INCREASE, CHANGE_DECREASE)
        NOCHANGE = CHANGE == False
    
    metrics = {}
    
    # These are all the possible definitions. Note: they *must* add up to one.
    metrics['deforestation'] = F_NF & CHANGE
    metrics['degradation'] = F_F & CHANGE & DECREASE
    metrics['minorloss'] = (F_F | F_NF) & NOCHANGE & DECREASE
    
    metrics['aforestation'] = NF_F & CHANGE
    metrics['growth'] = F_F & CHANGE & INCREASE
    metrics['minorgain'] = (F_F | NF_F) & NOCHANGE & INCREASE
            
    metrics['nonforest'] = NF_NF
    
    if output:
        output_im = np.zeros_like(AGB_t1) + data_t1.nodata
        output_im[metrics['nonforest'].data] = 0
        output_im[metrics['deforestation'].data] = 1
        output_im[metrics['degradation'].data] = 2
        output_im[metrics['minorloss'].data] = 3
        output_im[metrics['minorgain'].data] = 4
        output_im[metrics['growth'].data] = 5
        output_im[metrics['aforestation'].data] = 6
        
        output_im[np.logical_or(data_t1.mask, data_t2.mask)] = 99
        
        biota.IO.outputGeoTiff(output_im, data_t1.output_pattern%('CHANGE', '%s_%s'%(str(data_t1.year), str(data_t2.year))), data_t1.geo_t, data_t1.proj, output_dir = data_t1.output_dir, dtype = gdal.GDT_Byte, nodata = 99)
    
    # Also return AGB in metrics (with combined masks)
    metrics['AGB_t1'] = AGB_t1
    metrics['AGB_t2'] = AGB_t2
    
    return metrics




"""

def getChangeDownsampled(data_t1, data_t2, shrink_factor = 45, output = True):
    '''
    Downsample getChange to address false positives.
    
    This is a palceholder for something more advanced.
    '''
    
    print "WARNING: This function is not yet functional. Come back later!"
        
    metrics = getChange(data_t1, data_t2, F_threshold = 15., C_threshold = 0.25, min_area = 2.)
    
    # Calculate output output size based on reduction shrink_factor
    output_size = int(round(((data_t1.xSize + data_t1.ySize) / 2.) / shrink_factor))
    
    change_downsampled = np.zeros((output_size, output_size)) + data_t1.nodata
        
    for n, ymin in enumerate(np.linspace(0, data_t1.ySize, output_size, endpoint = False)):
        ymax = ymin + (data_t1.ySize / output_size)
        for m, xmin in enumerate(np.linspace(0, data_t1.xSize, output_size, endpoint = False)):
            xmax = xmin + (data_t1.xSize / output_size)
            
            deforestation = metrics['deforestation'][int(round(ymin)):int(round(ymax)), int(round(xmin)):int(round(xmax))]
            aforestation = metrics['aforestation'][int(round(ymin)):int(round(ymax)), int(round(xmin)):int(round(xmax))]
            this_mask = np.logical_or(data_t1.mask, data_t2.mask)[int(round(ymin)):int(round(ymax)), int(round(xmin)):int(round(xmax))]
            
            change_downsampled[n,m] = aforestation.sum() - deforestation.sum()
        
    if output: biota.IO.outputGeoTiff(change_downsampled, data_t1.output_pattern%('CHANGEDOWNSAMPLED', '%s_%s'%(str(data_t1.year), str(data_t2.year))), data_t1.shrinkGeoT(shrink_factor), data_t1.proj, output_dir = data_t1.output_dir, dtype = gdal.GDT_Float32, nodata = data_t1.nodata)
"""


from setuptools import setup

setup(name='biota',
      packages = ['biota'],
      version='0.2',
      description='Tools to generate aboveground biomass forest change maps from ALOS PALSAR/PALSAR-2 mosaic data.',
      url='https://bitbucket.org/sambowers/biota',
      author='Samuel Bowers',
      author_email='sam.bowers@ed.ac.uk',
      license='GNU General Public License',
      zip_safe=False)
